__author__ = "dangilman"

from lenstronomy.LensModel.MultiPlane.multi_plane import MultiPlane
from lenstronomy.LensModel.single_plane import SinglePlane
from lenstronomy.Cosmo.background import Background

__all__ = ["MultiPlaneDecoupled"]


class MultiPlaneDecoupled(MultiPlane):
    def __init__(
        self,
        z_source,
        lens_model_list,
        lens_redshift_list,
        cosmo=None,
        numerical_alpha_class=None,
        observed_convention_index=None,
        ignore_observed_positions=False,
        z_source_convention=None,
        cosmo_interp=False,
        z_interp_stop=None,
        num_z_interp=100,
        kwargs_interp=None,
        kwargs_synthesis=None,
        x0_interp=None,
        y0_interp=None,
        alpha_x_interp_foreground=None,
        alpha_y_interp_foreground=None,
        alpha_x_interp_background=None,
        alpha_y_interp_background=None,
        z_split=None,
    ):
        """A class for multiplane lensing in which the deflection angles at certain
        coordinates are fixed through user-specified interpolation functions. These
        functions return fixed deflection angles that effectively decouple deflections
        by a group of deflectors at redshift Z from deflections produced by halos at
        redshift< Z.

        This class breaks the recursive nature of the multi-plane lens equation, and can
        significantly speed up computations with a large number of line-of-sight halos.

        :param lens_model_list: list of lens model strings
        :param lens_redshift_list: list of floats with redshifts of the lens models
            indicated in lens_model_list
        :param z_source_convention: float, redshift of a source to define the reduced
            deflection angles of the lens models. If None, 'z_source' is used.
        :param cosmo: instance of astropy.cosmology
        :param numerical_alpha_class: an instance of a custom class for use in
            NumericalAlpha() lens model (see documentation in Profiles/numerical_alpha)
        :param kwargs_interp: interpolation keyword arguments specifying the numerics.
            See description in the Interpolate() class. Only applicable for 'INTERPOL'
            and 'INTERPOL_SCALED' models.
        :param kwargs_synthesis: keyword arguments for the 'SYNTHESIS' lens model, if
            applicable
        :param x0_interp: a function that maps an angular coordinate on the sky to the x
            coordinate of a physical position [Mpc] at the first lens plane
        :param y0_interp: same as x0_interp, but returns the y coordinate in Mpc
        :param alpha_x_interp_list: a list of functions that take as input angular
            coordinates (x, y) and returns the x-component of the deflection angle at
            each coorindate
        :param alpha_y_interp_list: same as alpha_x_interp_list, but returns the
            y-component of the deflection angle at (x,y)
        :param z_interp_list: a list of redshifts corresponding to the
            alpha_x_interp_list and alpha_y_interp_list entries
        """
        self._alphax_interp_foreground = alpha_x_interp_foreground
        self._alphay_interp_foreground = alpha_y_interp_foreground
        self._alphax_interp_background = alpha_x_interp_background
        self._alphay_interp_background = alpha_y_interp_background
        self._x0_interp = x0_interp
        self._y0_interp = y0_interp
        self._z_split = z_split
        super(MultiPlaneDecoupled, self).__init__(
            z_source,
            lens_model_list,
            lens_redshift_list,
            cosmo,
            numerical_alpha_class,
            observed_convention_index,
            ignore_observed_positions,
            z_source_convention,
            cosmo_interp,
            z_interp_stop,
            num_z_interp,
            kwargs_interp,
            kwargs_synthesis,
        )

        cosmo_bkg = Background(cosmo)
        d_xy_source = cosmo_bkg.d_xy(0, z_source)
        d_xy_lens_source = cosmo_bkg.d_xy(self._z_split, z_source)
        self._reduced_to_phys = d_xy_source / d_xy_lens_source
        self._deltaT_zsplit = cosmo_bkg.d_xy(0, self._z_split)
        self._deltaT_zsplit_zsource = cosmo_bkg.T_xy(self._z_split, z_source)
        self._Ts = cosmo_bkg.T_xy(0, z_source)
        self._Td = cosmo_bkg.T_xy(0, z_split)
        self._Tds = cosmo_bkg.T_xy(self._z_split, z_source)
        self._main_deflector = SinglePlane(lens_model_list)
        # useful to have these saved to access later outside the class
        self.kwargs_multiplane_model = {
            "x0_interp": self._x0_interp,
            "y0_interp": self._y0_interp,
            "alpha_x_interp_foreground": self._alphax_interp_foreground,
            "alpha_y_interp_foreground": self._alphay_interp_foreground,
            "alpha_x_interp_background": self._alphax_interp_background,
            "alpha_y_interp_background": self._alphay_interp_background,
            "z_split": self._z_split,
        }

    def geo_shapiro_delay(*args, **kwargs):
        raise Exception(
            "time delays are not yet implemented for the MultiPlaneDecoupled class"
        )

    def ray_shooting_partial_comoving(self, *args, **kwargs):
        raise Exception("ray_shooting_partial is not well defined for this class")

    def ray_shooting_partial_comoving(self, *args, **kwargs):
        raise Exception(
            "ray_shooting_partial_comoving is not well defined for this class"
        )

    def ray_shooting(self, theta_x, theta_y, kwargs_lens, *args, **kwargs):
        """Ray-shooting through the lens volume with fixed deflection angles at certain
        lens planes passed through the alpha_x_interp/alpha_y_interp lists. Starts with
        (x,y) co-moving distance passed through the x0_interp and y0_interp functions,
        then starts multi-plane ray-tracing through all subsequent lens planes.

        :param theta_x: angular coordinate on the sky
        :param theta_y: angular coordinate on the sky
        :param kwargs_lens: keyword arguments for the main deflector
        :return: coordinates on the source plane
        """
        coordinates = (theta_x, theta_y)
        # here we use an interpolation function to compute the comoving coordinates of the light rays
        # where they hit the main lens plane at redshift z = z_main
        x = self._x0_interp(coordinates)
        y = self._y0_interp(coordinates)

        theta_x_main = x / self._Td  # the angular coordinates of the ray positions
        theta_y_main = y / self._Td  # the angular coordinates of the ray positions

        # now we compute (via the interpolation functions) the deflection angles from all deflectors at z <= z_main, \
        # exlucding the main deflector
        angle_x_foreground = self._alphax_interp_foreground(coordinates)
        angle_y_foreground = self._alphay_interp_foreground(coordinates)

        # compute the deflection angles from the main deflector
        deflection_x_main, deflection_y_main = self._main_deflector.alpha(
            theta_x_main, theta_y_main, kwargs_lens
        )

        deflection_x_main *= self._reduced_to_phys
        deflection_y_main *= self._reduced_to_phys

        # add the main deflector to the deflection field
        angle_x = angle_x_foreground - deflection_x_main
        angle_y = angle_y_foreground - deflection_y_main

        # now we compute (via the interpolation functions) the deflection angles from all deflectors at z > z_main
        deflection_x_background = self._alphax_interp_background(coordinates)
        deflection_y_background = self._alphay_interp_background(coordinates)

        # combine deflections
        alpha_background_x = angle_x + deflection_x_background
        alpha_background_y = angle_y + deflection_y_background

        # ray propagation to the source plane with the small angle approximation
        beta_x = x / self._Ts + alpha_background_x * self._Tds / self._Ts
        beta_y = y / self._Ts + alpha_background_y * self._Tds / self._Ts

        return beta_x, beta_y



from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.ImSim.image2source_mapping import Image2SourceMapping
from lenstronomy.Util import util
import numpy as np


class TracerModelSource(ImageModel):
    """Tracer model class, inherits ImageModel."""

    def __init__(
        self,
        data_class,
        psf_class=None,
        lens_model_class=None,
        source_model_class=None,
        lens_light_model_class=None,
        point_source_class=None,
        extinction_class=None,
        tracer_source_class=None,
        kwargs_numerics=None,
        likelihood_mask=None,
        psf_error_map_bool_list=None,
        kwargs_pixelbased=None,
        tracer_partition=None,
        tracer_type="LINEAR",
    ):
        """

        :param data_class: ImageData() instance
        :param psf_class: PSF() instance
        :param lens_model_class: LensModel() instance
        :param source_model_class: LightModel() instance
        :param lens_light_model_class: LightModel() instance
        :param point_source_class: PointSource() instance
        :param tracer_source_class: LightModel() instance describing the tracers of the source
        :param kwargs_numerics: keyword arguments passed to the Numerics module
        :param likelihood_mask: 2d boolean array of pixels to be counted in the likelihood calculation/linear
         optimization
        :param psf_error_map_bool_list: list of boolean of length of point source models.
         Indicates whether PSF error map is used for the point source model stated as the index.
        :param kwargs_pixelbased: keyword arguments with various settings related to the pixel-based solver
         (see SLITronomy documentation) being applied to the point sources.
        :param tracer_partition: in case of tracer models for specific sub-parts of the surface brightness model
         [[list of light profiles, list of tracer profiles], [list of light profiles, list of tracer profiles], [...], ...]
        :type tracer_partition: None or list
        :param tracer_type: LINEAR or LOG. If the tracer is in log units, it is converted to linear units, summed,
         and then converted back to log units.
        :type tracer_partition: str
        """
        if likelihood_mask is None:
            likelihood_mask = np.ones_like(data_class.data)
        self.likelihood_mask = np.array(likelihood_mask, dtype=bool)
        self._mask1d = util.image2array(self.likelihood_mask)
        if tracer_partition is None:
            tracer_partition = [[None, None]]
        self._tracer_partition = tracer_partition
        if tracer_type not in ["LINEAR", "LOG"]:
            raise Exception(
                "Unknown input tracer_type: {0}. Only two tracer types are currently supported: LINEAR and LOG. Please convert your tracer to linear/log units.".format(
                    tracer_type
                )
            )
        self._tracer_type = tracer_type
        super(TracerModelSource, self).__init__(
            data_class,
            psf_class=psf_class,
            lens_model_class=lens_model_class,
            source_model_class=source_model_class,
            lens_light_model_class=lens_light_model_class,
            point_source_class=point_source_class,
            extinction_class=extinction_class,
            kwargs_numerics=kwargs_numerics,
            kwargs_pixelbased=kwargs_pixelbased,
        )
        if psf_error_map_bool_list is None:
            psf_error_map_bool_list = [True] * len(
                self.PointSource.point_source_type_list
            )
        self._psf_error_map_bool_list = psf_error_map_bool_list
        if tracer_source_class is None:
            tracer_source_class = LightModel(light_model_list=[])
        if lens_model_class is None:
            lens_model_class = LensModel(lens_model_list=[])
        self.tracer_mapping = Image2SourceMapping(
            lens_model=lens_model_class, source_model=tracer_source_class
        )
        self.tracer_source_class = tracer_source_class

    def tracer_model(
        self,
        kwargs_tracer_source,
        kwargs_lens,
        kwargs_source,
        kwargs_extinction=None,
        kwargs_special=None,
        de_lensed=False,
    ):
        """Tracer model as a convolved surface brightness weighted quantity conv(tracer
        * surface brightness) / conv(surface brightness)

        :param kwargs_tracer_source:
        :param kwargs_lens:
        :param kwargs_source:
        :return: model predicted observed tracer component
        """
        tracer_brightness_conv = np.zeros_like(self.Data.data)
        source_light_conv = np.zeros_like(self.Data.data)
        for [k_light, k_tracer] in self._tracer_partition:
            source_light_k = self._source_surface_brightness_analytical_numerics(
                kwargs_source,
                kwargs_lens,
                kwargs_extinction,
                kwargs_special=kwargs_special,
                de_lensed=de_lensed,
                k=k_light,
            )
            source_light_conv_k = self.ImageNumerics.re_size_convolve(
                source_light_k, unconvolved=False
            )
            source_light_conv_k[source_light_conv_k < 10 ** (-20)] = 10 ** (-20)
            tracer_k = self._tracer_model_source(
                kwargs_tracer_source, kwargs_lens, de_lensed=de_lensed, k=k_tracer
            )
            if self._tracer_type == "LINEAR":
                tracer_brightness_conv_k = self.ImageNumerics.re_size_convolve(
                    tracer_k * source_light_k, unconvolved=False
                )
                tracer_brightness_conv += tracer_brightness_conv_k
            if self._tracer_type == "LOG":
                lin_tracer_k = 10 ** (tracer_k)
                lin_tracer_brightness_conv_k = self.ImageNumerics.re_size_convolve(
                    lin_tracer_k * source_light_k, unconvolved=False
                )
                tracer_brightness_conv += lin_tracer_brightness_conv_k

            source_light_conv += source_light_conv_k
        if self._tracer_type == "LINEAR":
            return tracer_brightness_conv / source_light_conv
        if self._tracer_type == "LOG":
            return np.log10(tracer_brightness_conv / source_light_conv)

    def _tracer_model_source(
        self, kwargs_tracer_source, kwargs_lens, de_lensed=False, k=None
    ):
        """

        :param kwargs_tracer_source:
        :param kwargs_lens:
        :return:
        """
        ra_grid, dec_grid = self.ImageNumerics.coordinates_evaluate
        if de_lensed is True:
            source_light = self.tracer_source_class.surface_brightness(
                ra_grid, dec_grid, kwargs_tracer_source, k=k
            )
        else:
            source_light = self.tracer_mapping.image_flux_joint(
                ra_grid, dec_grid, kwargs_lens, kwargs_tracer_source, k=k
            )
        return source_light

    def likelihood_data_given_model(
        self,
        kwargs_tracer_source,
        kwargs_lens,
        kwargs_source,
        kwargs_extinction=None,
        kwargs_special=None,
    ):
        model = self.tracer_model(
            kwargs_tracer_source,
            kwargs_lens,
            kwargs_source,
            kwargs_extinction,
            kwargs_special,
        )
        log_likelihood = self.Data.log_likelihood(
            model, self.likelihood_mask, additional_error_map=0
        )
        return log_likelihood

    @property
    def num_data_evaluate(self):
        """Number of data points to be used in the linear solver :return:"""
        return int(np.sum(self.likelihood_mask))



__author__ = "dangilman"

import numpy.testing as npt
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Util.decouple_multi_plane_util import (
    setup_lens_model,
    setup_grids,
    coordinates_and_deflections,
    setup_raytracing_lensmodels,
    decoupled_multiplane_class_setup,
)
from copy import deepcopy
import numpy as np
import pytest


class TestMultiPlaneDecoupled(object):

    def setup_method(self):
        self.zlens = 0.5
        self.z_source = 2.0
        self.kwargs_lens_true = [
            {
                "theta_E": 0.7,
                "center_x": 0.0,
                "center_y": -0.0,
                "e1": 0.2,
                "e2": -0.1,
                "gamma": 2.0,
            },
            {"theta_E": 0.2, "center_x": 0.0, "center_y": -0.4},
            {"theta_E": 0.2, "center_x": 0.6, "center_y": 0.3},
            {"theta_E": 0.15, "center_x": -0.6, "center_y": -1.0},
            {"gamma1": 0.1, "gamma2": -0.2},
        ]
        self.lens_model_list = ["EPL", "SIS", "SIS", "SIS", "SHEAR"]
        self.index_lens_split = [0, 1]
        self.lens_redshift_list = [self.zlens, self.zlens, 0.25, 1.0, 1.5]
        self.lens_model_true = LensModel(
            self.lens_model_list,
            lens_redshift_list=self.lens_redshift_list,
            multi_plane=True,
            z_source=self.z_source,
        )
        self.cosmo = self.lens_model_true.cosmo

        (
            self.lens_model_fixed,
            self.lens_model_free,
            self.kwargs_lens_fixed,
            self.kwargs_lens_free,
            self.z_source,
            self.z_split,
            self.cosmo_bkg,
        ) = setup_lens_model(
            self.lens_model_true, self.kwargs_lens_true, self.index_lens_split
        )

        self.Td = self.cosmo_bkg.T_xy(0, self.zlens)
        self.Ts = self.cosmo_bkg.T_xy(0, self.z_source)
        self.Tds = self.cosmo_bkg.T_xy(self.zlens, self.z_source)
        self.reduced_to_phys = self.cosmo_bkg.d_xy(
            0, self.z_source
        ) / self.cosmo_bkg.d_xy(self.zlens, self.z_source)

        # SETUP TESTS FOR POINT, GRID, AND MULTIPLE IMAGES
        self.x_image = [1.0, 0.5, -0.5, -1.0]
        self.y_image = [-0.4, 0.5, 1.0, 0.7]
        self.x_point = self.x_image[1]
        self.y_point = self.y_image[1]
        self._setup_point()
        self._setup_grid()
        self._setup_multiple_images()

    def _setup_point(self):
        # SETUP TESTS FOR POINT, GRID, AND MULTIPLE IMAGES

        (
            self.x0_point,
            self.y0_point,
            self.alphax_foreground_point,
            self.alphay_foreground_point,
            self.alphax_background_point,
            self.alphay_background_point,
        ) = coordinates_and_deflections(
            self.lens_model_fixed,
            self.lens_model_free,
            self.kwargs_lens_fixed,
            self.kwargs_lens_free,
            self.x_image[1],
            self.y_image[1],
            self.z_split,
            self.z_source,
            self.cosmo_bkg,
        )

        self.kwargs_multiplane_model_point = decoupled_multiplane_class_setup(
            self.lens_model_free,
            self.x0_point,
            self.y0_point,
            self.alphax_foreground_point,
            self.alphay_foreground_point,
            self.alphax_background_point,
            self.alphay_background_point,
            self.z_split,
            coordinate_type="POINT",
        )

    def _setup_grid(self):
        # SETUP TESTS FOR POINT, GRID, AND MULTIPLE IMAGES
        grid_size = 2.5
        grid_resolution = 0.005
        self.grid_x, self.grid_y, self.interp_points_grid, self.npix_grid = setup_grids(
            grid_size, grid_resolution, 0.0, 0.0
        )
        (
            self.x0_grid,
            self.y0_grid,
            self.alphax_foreground_grid,
            self.alphay_foreground_grid,
            self.alphax_background_grid,
            self.alphay_background_grid,
        ) = coordinates_and_deflections(
            self.lens_model_fixed,
            self.lens_model_free,
            self.kwargs_lens_fixed,
            self.kwargs_lens_free,
            self.grid_x,
            self.grid_y,
            self.z_split,
            self.z_source,
            self.cosmo_bkg,
        )

        self.kwargs_multiplane_model_grid = decoupled_multiplane_class_setup(
            self.lens_model_free,
            self.x0_grid,
            self.y0_grid,
            self.alphax_foreground_grid,
            self.alphay_foreground_grid,
            self.alphax_background_grid,
            self.alphay_background_grid,
            self.z_split,
            coordinate_type="GRID",
            interp_points=self.interp_points_grid,
        )

    def _setup_multiple_images(self):
        # SETUP TESTS FOR POINT, GRID, AND MULTIPLE IMAGES

        (
            self.x0_MI,
            self.y0_MI,
            self.alphax_foreground_MI,
            self.alphay_foreground_MI,
            self.alphax_background_MI,
            self.alphay_background_MI,
        ) = coordinates_and_deflections(
            self.lens_model_fixed,
            self.lens_model_free,
            self.kwargs_lens_fixed,
            self.kwargs_lens_free,
            self.x_image,
            self.y_image,
            self.z_split,
            self.z_source,
            self.cosmo_bkg,
        )

        self.kwargs_multiplane_model_MI = decoupled_multiplane_class_setup(
            self.lens_model_free,
            self.x0_MI,
            self.y0_MI,
            self.alphax_foreground_MI,
            self.alphay_foreground_MI,
            self.alphax_background_MI,
            self.alphay_background_MI,
            self.z_split,
            coordinate_type="MULTIPLE_IMAGES",
            x_image=self.x_image,
            y_image=self.y_image,
        )

    def test_zsplit_requirements(self):
        """
        TODO: generalize to an arbitrary set of deflectors at different redshifts with deflection angles decoupled from
        the line of sight
        """
        index_lens_split = [0, 1, 2]
        args = (self.lens_model_true, self.kwargs_lens_true, index_lens_split)
        npt.assert_raises(Exception, setup_lens_model, args)

    def test_setup_grids(self):
        grid_size = 2.0
        grid_resolution = 0.001
        ximg = 0.5
        yimg = -0.2
        xx, yy, interp_points, npixels = setup_grids(
            grid_size,
            grid_resolution,
            coordinate_center_x=ximg,
            coordinate_center_y=yimg,
        )
        npt.assert_equal(npixels - 1, grid_size / grid_resolution)
        npt.assert_equal(xx[0], -1 + ximg)
        npt.assert_equal(yy[0], -1 + yimg)
        npt.assert_equal(interp_points[0][0], -1 + ximg)
        npt.assert_equal(interp_points[1][0], -1 + yimg)
        npt.assert_equal(interp_points[0][-1], 1 + ximg)
        npt.assert_equal(interp_points[1][-1], 1 + yimg)

    def test_setup_lens_model(self):

        (
            lens_model_fixed,
            lens_model_free,
            kwargs_lens_fixed,
            kwargs_lens_free,
            z_source,
            z_split,
            cosmo_bkg,
        ) = setup_lens_model(
            self.lens_model_true, self.kwargs_lens_true, self.index_lens_split
        )
        npt.assert_equal(len(lens_model_fixed.lens_model_list), 3)
        npt.assert_equal(
            len(lens_model_free.lens_model_list), len(self.index_lens_split)
        )
        npt.assert_equal(lens_model_fixed.redshift_list[0], 0.25)
        npt.assert_equal(lens_model_fixed.redshift_list[1], 1.0)
        npt.assert_equal(lens_model_free.redshift_list[0], self.zlens)
        npt.assert_equal(lens_model_free.redshift_list[0], self.zlens)
        npt.assert_equal(z_source, 2.0)
        npt.assert_equal(z_split, self.zlens)
        npt.assert_equal(kwargs_lens_free[0]["theta_E"], 0.7)
        npt.assert_equal(kwargs_lens_free[1]["theta_E"], 0.2)
        npt.assert_equal(kwargs_lens_fixed[0]["theta_E"], 0.2)
        npt.assert_equal(kwargs_lens_fixed[1]["theta_E"], 0.15)

    def test_point_deflection_model(self):

        # the true source coordinate
        x_source_true, y_source_true, _, _ = (
            self.lens_model_true.lens_model.ray_shooting_partial_comoving(
                0.0,
                0.0,
                self.x_point,
                self.y_point,
                0.0,
                self.z_source,
                self.kwargs_lens_true,
            )
        )
        beta_x_true, beta_y_true = x_source_true / self.Ts, y_source_true / self.Ts
        npt.assert_almost_equal(
            beta_x_true,
            self.lens_model_true.ray_shooting(
                self.x_point, self.y_point, self.kwargs_lens_true
            )[0],
        )

        # create new dictionary with the macromodel effectively removed
        kwargs_lens_free_no_macromodel = deepcopy(self.kwargs_lens_true)
        kwargs_lens_free_no_macromodel[0]["theta_E"] = 0.0
        kwargs_lens_free_no_macromodel[1]["theta_E"] = 0.0

        # show that the x0, y0, and ray angles are as expected
        x0_true, y0_true, alphax_fore_true, alphay_fore_true = (
            self.lens_model_true.lens_model.ray_shooting_partial_comoving(
                0.0,
                0.0,
                self.x_point,
                self.y_point,
                0.0,
                self.zlens,
                kwargs_lens_free_no_macromodel,
            )
        )
        npt.assert_almost_equal(self.x0_point, x0_true)
        npt.assert_almost_equal(self.y0_point, y0_true)
        npt.assert_almost_equal(
            self.kwargs_multiplane_model_point["kwargs_multiplane_model"]["x0_interp"](
                self.x_point, self.y_point
            ),
            x0_true,
        )
        npt.assert_almost_equal(
            self.kwargs_multiplane_model_point["kwargs_multiplane_model"]["y0_interp"](
                self.x_point, self.y_point
            ),
            y0_true,
        )
        npt.assert_almost_equal(self.alphax_foreground_point, alphax_fore_true)
        npt.assert_almost_equal(self.alphay_foreground_point, alphay_fore_true)

        # compute the true ray angles at the main lens plane
        x0_true, y0_true, alphax_fore_true, alphay_fore_true = (
            self.lens_model_true.lens_model.ray_shooting_partial_comoving(
                0.0,
                0.0,
                self.x_point,
                self.y_point,
                0.0,
                self.zlens,
                self.kwargs_lens_true,
            )
        )
        npt.assert_almost_equal(self.x0_point, x0_true)
        npt.assert_almost_equal(self.y0_point, y0_true)

        # compute deflections from the macromodel only
        theta_x, theta_y = self.x0_point / self.Td, self.y0_point / self.Td
        alpha_x_macro, alpha_y_macro = 0.0, 0.0
        func_list = self.lens_model_true.lens_model._multi_plane_base.func_list
        for k in range(0, 2):
            _alpha_x_macromodel, _alpha_y_macromodel = func_list[k].derivatives(
                theta_x, theta_y, **self.kwargs_lens_true[k]
            )
            alpha_x_macro += _alpha_x_macromodel * self.reduced_to_phys
            alpha_y_macro += _alpha_y_macromodel * self.reduced_to_phys

        # the full deflection field at the main lens plane is alpha_x_true = alpha_x_foreground - alpha_x_macro
        # therefore alpha_x_foreground = alpha_x_true + alpha_x_macro
        npt.assert_almost_equal(
            self.alphax_foreground_point, alphax_fore_true + alpha_x_macro
        )
        npt.assert_almost_equal(
            self.alphay_foreground_point, alphay_fore_true + alpha_y_macro
        )
        npt.assert_almost_equal(
            self.kwargs_multiplane_model_point["kwargs_multiplane_model"][
                "alpha_x_interp_foreground"
            ](self.x_point, self.y_point),
            alphax_fore_true + alpha_x_macro,
        )
        npt.assert_almost_equal(
            self.kwargs_multiplane_model_point["kwargs_multiplane_model"][
                "alpha_y_interp_foreground"
            ](self.x_point, self.y_point),
            alphay_fore_true + alpha_y_macro,
        )

        # ray propagation with the small angle approx
        alpha_x = (
            self.alphax_foreground_point - alpha_x_macro + self.alphax_background_point
        )
        alpha_y = (
            self.alphay_foreground_point - alpha_y_macro + self.alphay_background_point
        )
        x_source_point = self.x0_point + alpha_x * self.Tds
        y_source_point = self.y0_point + alpha_y * self.Tds

        npt.assert_almost_equal(
            self.kwargs_multiplane_model_point["kwargs_multiplane_model"][
                "alpha_x_interp_background"
            ](self.x_point, self.y_point),
            self.alphax_background_point,
        )
        npt.assert_almost_equal(
            self.kwargs_multiplane_model_point["kwargs_multiplane_model"][
                "alpha_y_interp_background"
            ](self.x_point, self.y_point),
            self.alphay_background_point,
        )

        npt.assert_almost_equal(x_source_true, x_source_point)
        npt.assert_almost_equal(y_source_true, y_source_point)

        lens_model_decoupled = LensModel(**self.kwargs_multiplane_model_point)
        beta_x_new, beta_y_new = lens_model_decoupled.ray_shooting(
            self.x_image[1], self.y_image[1], self.kwargs_lens_free
        )
        beta_x_true, beta_y_true = self.lens_model_true.ray_shooting(
            self.x_image[1], self.y_image[1], self.kwargs_lens_true
        )
        npt.assert_allclose(beta_x_new, beta_x_true)
        npt.assert_allclose(beta_y_new, beta_y_true)

    def test_grid_deflection_model(self):

        # the true source coordinate
        x_source_true, y_source_true, _, _ = (
            self.lens_model_true.lens_model.ray_shooting_partial_comoving(
                np.zeros_like(self.grid_x),
                np.zeros_like(self.grid_y),
                self.grid_x,
                self.grid_y,
                0.0,
                self.z_source,
                self.kwargs_lens_true,
            )
        )

        # create new dictionary with the macromodel effectively removed
        kwargs_lens_free_no_macromodel = deepcopy(self.kwargs_lens_true)
        kwargs_lens_free_no_macromodel[0]["theta_E"] = 0.0
        kwargs_lens_free_no_macromodel[1]["theta_E"] = 0.0

        # show that the x0, y0, and ray angles are as expected
        x0_true, y0_true, alphax_fore_true, alphay_fore_true = (
            self.lens_model_true.lens_model.ray_shooting_partial_comoving(
                np.zeros_like(self.grid_x),
                np.zeros_like(self.grid_y),
                self.grid_x,
                self.grid_y,
                0.0,
                self.zlens,
                kwargs_lens_free_no_macromodel,
            )
        )
        npt.assert_almost_equal(self.x0_grid, x0_true)
        npt.assert_almost_equal(self.y0_grid, y0_true)
        npt.assert_almost_equal(self.alphax_foreground_grid, alphax_fore_true)
        npt.assert_almost_equal(self.alphay_foreground_grid, alphay_fore_true)

        # compute the true ray angles at the main lens plane
        x0_true, y0_true, alphax_fore_true, alphay_fore_true = (
            self.lens_model_true.lens_model.ray_shooting_partial_comoving(
                np.zeros_like(self.grid_x),
                np.zeros_like(self.grid_y),
                self.grid_x,
                self.grid_y,
                0.0,
                self.zlens,
                self.kwargs_lens_true,
            )
        )
        npt.assert_almost_equal(self.x0_grid, x0_true)
        npt.assert_almost_equal(self.y0_grid, y0_true)

        # compute deflections from the macromodel only
        theta_x, theta_y = self.x0_grid / self.Td, self.y0_grid / self.Td
        alpha_x_macro, alpha_y_macro = 0.0, 0.0
        func_list = self.lens_model_true.lens_model._multi_plane_base.func_list
        for k in range(0, 2):
            _alpha_x_macromodel, _alpha_y_macromodel = func_list[k].derivatives(
                theta_x, theta_y, **self.kwargs_lens_true[k]
            )
            alpha_x_macro += _alpha_x_macromodel * self.reduced_to_phys
            alpha_y_macro += _alpha_y_macromodel * self.reduced_to_phys

        # the full deflection field at the main lens plane is alpha_x_true = alpha_x_foreground - alpha_x_macro
        # therefore alpha_x_foreground = alpha_x_true + alpha_x_macro
        npt.assert_almost_equal(
            self.alphax_foreground_grid, alphax_fore_true + alpha_x_macro
        )
        npt.assert_almost_equal(
            self.alphay_foreground_grid, alphay_fore_true + alpha_y_macro
        )

        # ray propagation with the small angle approx
        alpha_x = (
            self.alphax_foreground_grid - alpha_x_macro + self.alphax_background_grid
        )
        alpha_y = (
            self.alphay_foreground_grid - alpha_y_macro + self.alphay_background_grid
        )
        x_source_grid = self.x0_grid + alpha_x * self.Tds
        y_source_grid = self.y0_grid + alpha_y * self.Tds

        npt.assert_almost_equal(x_source_true, x_source_grid)
        npt.assert_almost_equal(y_source_true, y_source_grid)

        lens_model_decoupled = LensModel(**self.kwargs_multiplane_model_grid)
        beta_x_new, beta_y_new = lens_model_decoupled.ray_shooting(
            self.x_point, self.y_point, self.kwargs_lens_free
        )
        beta_x_true, beta_y_true = self.lens_model_true.ray_shooting(
            self.x_point, self.y_point, self.kwargs_lens_true
        )
        npt.assert_almost_equal(beta_x_new, beta_x_true)
        npt.assert_almost_equal(beta_y_new, beta_y_true)

    def test_multiple_images_deflection_model(self):

        # the true source coordinate
        x_source_true, y_source_true, _, _ = (
            self.lens_model_true.lens_model.ray_shooting_partial_comoving(
                np.zeros_like(self.x_image),
                np.zeros_like(self.y_image),
                self.x_image,
                self.y_image,
                0.0,
                self.z_source,
                self.kwargs_lens_true,
            )
        )

        # create new dictionary with the macromodel effectively removed
        kwargs_lens_free_no_macromodel = deepcopy(self.kwargs_lens_true)
        kwargs_lens_free_no_macromodel[0]["theta_E"] = 0.0
        kwargs_lens_free_no_macromodel[1]["theta_E"] = 0.0

        # show that the x0, y0, and ray angles are as expected
        x0_true, y0_true, alphax_fore_true, alphay_fore_true = (
            self.lens_model_true.lens_model.ray_shooting_partial_comoving(
                np.zeros_like(self.y_image),
                np.zeros_like(self.x_image),
                self.x_image,
                self.y_image,
                0.0,
                self.zlens,
                kwargs_lens_free_no_macromodel,
            )
        )
        npt.assert_almost_equal(self.x0_MI, x0_true)
        npt.assert_almost_equal(self.y0_MI, y0_true)
        npt.assert_almost_equal(self.alphax_foreground_MI, alphax_fore_true)
        npt.assert_almost_equal(self.alphay_foreground_MI, alphay_fore_true)

        # compute the true ray angles at the main lens plane
        x0_true, y0_true, alphax_fore_true, alphay_fore_true = (
            self.lens_model_true.lens_model.ray_shooting_partial_comoving(
                np.zeros_like(self.y_image),
                np.zeros_like(self.x_image),
                self.x_image,
                self.y_image,
                0.0,
                self.zlens,
                self.kwargs_lens_true,
            )
        )
        npt.assert_almost_equal(self.x0_MI, x0_true)
        npt.assert_almost_equal(self.y0_MI, y0_true)

        # compute deflections from the macromodel only
        theta_x, theta_y = self.x0_MI / self.Td, self.y0_MI / self.Td
        alpha_x_macro, alpha_y_macro = 0.0, 0.0
        func_list = self.lens_model_true.lens_model._multi_plane_base.func_list
        for k in range(0, 2):
            _alpha_x_macromodel, _alpha_y_macromodel = func_list[k].derivatives(
                theta_x, theta_y, **self.kwargs_lens_true[k]
            )
            alpha_x_macro += _alpha_x_macromodel * self.reduced_to_phys
            alpha_y_macro += _alpha_y_macromodel * self.reduced_to_phys

        # the full deflection field at the main lens plane is alpha_x_true = alpha_x_foreground - alpha_x_macro
        # therefore alpha_x_foreground = alpha_x_true + alpha_x_macro
        npt.assert_almost_equal(
            self.alphax_foreground_MI, alphax_fore_true + alpha_x_macro
        )
        npt.assert_almost_equal(
            self.alphay_foreground_MI, alphay_fore_true + alpha_y_macro
        )

        # ray propagation with the small angle approx
        alpha_x = self.alphax_foreground_MI - alpha_x_macro + self.alphax_background_MI
        alpha_y = self.alphay_foreground_MI - alpha_y_macro + self.alphay_background_MI
        x_source_grid = self.x0_MI + alpha_x * self.Tds
        y_source_grid = self.y0_MI + alpha_y * self.Tds

        npt.assert_allclose(x_source_true, x_source_grid)
        npt.assert_allclose(y_source_true, y_source_grid)

        lens_model_decoupled = LensModel(**self.kwargs_multiplane_model_MI)
        beta_x_new, beta_y_new = lens_model_decoupled.ray_shooting(
            self.x_image, self.y_image, self.kwargs_lens_free
        )
        beta_x_true, beta_y_true = self.lens_model_true.ray_shooting(
            self.x_image, self.y_image, self.kwargs_lens_true
        )
        npt.assert_allclose(beta_x_new, beta_x_true)
        npt.assert_allclose(beta_y_new, beta_y_true)

    def test_setup_raytracing_lens_models(self):

        grid_size = 0.05
        grid_resolution = 0.001
        multiplane_lens_model_list, kwargs_multiplane_lens_model_list = (
            setup_raytracing_lensmodels(
                self.x_image,
                self.y_image,
                self.lens_model_true,
                self.kwargs_lens_true,
                self.index_lens_split,
                grid_size,
                grid_resolution,
            )
        )
        x_grid, y_grid, _, _ = setup_grids(grid_size, 0.0025)
        lens_model_decoupled_multiple_images = LensModel(
            **self.kwargs_multiplane_model_MI
        )

        for i in range(0, len(self.x_image)):
            beta_x, beta_y = multiplane_lens_model_list[i].ray_shooting(
                x_grid + self.x_image[i],
                y_grid + self.y_image[i],
                self.kwargs_lens_free,
            )
            beta_x_true, beta_y_true = self.lens_model_true.ray_shooting(
                x_grid + self.x_image[i],
                y_grid + self.y_image[i],
                self.kwargs_lens_true,
            )
            npt.assert_allclose(beta_x, beta_x_true, 5)
            npt.assert_allclose(beta_y, beta_y_true, 5)
            beta_x, beta_y = multiplane_lens_model_list[i].ray_shooting(
                self.x_image[i], self.y_image[i], self.kwargs_lens_free
            )
            beta_x_true, beta_y_true = (
                lens_model_decoupled_multiple_images.ray_shooting(
                    x_grid + self.x_image[i],
                    y_grid + self.y_image[i],
                    self.kwargs_lens_free,
                )
            )
            npt.assert_almost_equal(beta_x, beta_x_true)
            npt.assert_almost_equal(beta_y, beta_y_true)


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



__author__ = "lucateo"

from lenstronomy.LensModel.Profiles.cored_density_exp import CoredDensityExp

import numpy as np
import numpy.testing as npt
import pytest


class TestCoredDensityExp(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.model = CoredDensityExp()

    def test_function(self):
        r = np.linspace(start=0.01, stop=2, num=10)
        kappa_0 = 0.1
        theta_c = 5
        f_ = self.model.function(r, 0, kappa_0, theta_c)
        delta = 0.0001
        f_d = self.model.function(r + delta, 0, kappa_0, theta_c)
        f_x_num = (f_d - f_) / delta
        f_x, _ = self.model.derivatives(r, 0, kappa_0, theta_c)
        npt.assert_almost_equal(f_x_num, f_x, decimal=3)
        # Try MSD limit
        theta_c_large = 13
        f_reference = self.model.function(0, 0, kappa_0, theta_c_large)
        f_large = self.model.function(r, 0, kappa_0, theta_c_large)
        f_MSD = 0.5 * kappa_0 * r**2
        npt.assert_almost_equal(f_large - f_reference, f_MSD, decimal=3)

    def test_derivatives(self):
        x = 0.5
        y = 0.8
        r = np.sqrt(x**2 + y**2)
        kappa_0, theta_c = 0.2, 9  # Trying MSD limit
        f_x, f_y = self.model.derivatives(x, y, kappa_0, theta_c)
        alpha_MSD = kappa_0 * r
        npt.assert_almost_equal(f_x, alpha_MSD * x / r, decimal=3)
        npt.assert_almost_equal(f_y, alpha_MSD * y / r, decimal=3)

    def test_hessian(self):
        x = np.linspace(start=0.01, stop=100, num=100)
        y = 0
        r = np.sqrt(x**2 + y**2)
        kappa_0 = 0.12
        theta_c = 6
        f_xx, f_xy, f_yx, f_yy = self.model.hessian(x, y, kappa_0, theta_c)
        kappa = 1.0 / 2 * (f_xx + f_yy)
        kappa_direct = self.model.kappa_r(r, kappa_0, theta_c)
        npt.assert_almost_equal(kappa, kappa_direct, decimal=5)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)

    def test_mass_3d(self):
        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        r = np.sqrt(x**2 + y**2)
        kappa_0 = 0.1
        theta_c = 7
        m3d = self.model.mass_3d(r, kappa_0, theta_c)
        m3d_lens = self.model.mass_3d_lens(r, kappa_0, theta_c)
        npt.assert_almost_equal(m3d, m3d_lens, decimal=8)


if __name__ == "__main__":
    pytest.main()



"""Provisional LSST instrument and observational settings.

See Optics and Observation Conditions spreadsheet at
https://docs.google.com/spreadsheets/d/1pMUB_OOZWwXON2dd5oP8PekhCT5MBBZJO1HV7IMZg4Y/edit?usp=sharing
for list of
sources.
"""

import copy
import lenstronomy.Util.util as util

__all__ = ["LSST"]

u_band_obs = {
    "exposure_time": 15.0,
    "sky_brightness": 22.99,
    "magnitude_zero_point": 26.5,
    "num_exposures": 140,
    "seeing": 0.81,
    "psf_type": "GAUSSIAN",
}

g_band_obs = {
    "exposure_time": 15.0,
    "sky_brightness": 22.26,
    "magnitude_zero_point": 28.30,
    "num_exposures": 200,
    "seeing": 0.77,
    "psf_type": "GAUSSIAN",
}

r_band_obs = {
    "exposure_time": 15.0,
    "sky_brightness": 21.2,
    "magnitude_zero_point": 28.13,
    "num_exposures": 460,
    "seeing": 0.73,
    "psf_type": "GAUSSIAN",
}

i_band_obs = {
    "exposure_time": 15.0,
    "sky_brightness": 20.48,
    "magnitude_zero_point": 27.79,
    "num_exposures": 460,
    "seeing": 0.71,
    "psf_type": "GAUSSIAN",
}

z_band_obs = {
    "exposure_time": 15.0,
    "sky_brightness": 19.6,
    "magnitude_zero_point": 27.40,
    "num_exposures": 400,
    "seeing": 0.69,
    "psf_type": "GAUSSIAN",
}

y_band_obs = {
    "exposure_time": 15.0,
    "sky_brightness": 18.61,
    "magnitude_zero_point": 26.58,
    "num_exposures": 400,
    "seeing": 0.68,
    "psf_type": "GAUSSIAN",
}

# - keyword exposure_time: exposure time per image (in seconds)
# - keyword sky_brightness: sky brightness (in magnitude per square arcseconds in units of electrons)
# - keyword magnitude_zero_point: magnitude in which 1 count (e-) per second per arcsecond square is registered
# - keyword num_exposures: number of exposures that are combined (depends on coadd_years)
#     when coadd_years = 10: num_exposures is baseline num of visits over 10 years (x2 since 2x15s exposures per visit)
# - keyword seeing: Full-Width-at-Half-Maximum (FWHM) of PSF
# - keyword psf_type: string, type of PSF ('GAUSSIAN' supported)


class LSST(object):
    """Class contains LSST instrument and observation configurations."""

    def __init__(self, band="g", psf_type="GAUSSIAN", coadd_years=10):
        """

        :param band: string, 'u', 'g', 'r', 'i', 'z' or 'y' supported. Determines obs dictionary.
        :param psf_type: string, type of PSF ('GAUSSIAN' supported).
        :param coadd_years: int, number of years corresponding to num_exposures in obs dict. Currently supported: 1-10.
        """
        if band.isalpha():
            band = band.lower()
        if band == "g":
            self.obs = copy.deepcopy(g_band_obs)
        elif band == "r":
            self.obs = copy.deepcopy(r_band_obs)
        elif band == "i":
            self.obs = copy.deepcopy(i_band_obs)
        elif band == "u":
            self.obs = copy.deepcopy(u_band_obs)
        elif band == "z":
            self.obs = copy.deepcopy(z_band_obs)
        elif band == "y":
            self.obs = copy.deepcopy(y_band_obs)
        else:
            raise ValueError(
                "band %s not supported! Choose 'u', 'g', 'r', 'i', 'z' or 'y'." % band
            )

        if psf_type != "GAUSSIAN":
            raise ValueError("psf_type %s not supported!" % psf_type)

        if coadd_years > 10 or coadd_years < 1:
            raise ValueError(
                " %s coadd_years not supported! Choose an integer between 1 and 10."
                % coadd_years
            )
        elif coadd_years != 10:
            self.obs["num_exposures"] = coadd_years * self.obs["num_exposures"] // 10

        self.camera = {
            "read_noise": 10,  # will be <10
            "pixel_scale": 0.2,
            "ccd_gain": 2.3,
        }

        # - keyword read_noise: std of noise generated by read-out (in units of electrons)
        # - keyword pixel_scale: scale (in arcseconds) of pixels
        # - keyword ccd_gain: electrons/ADU (analog-to-digital unit)

    def kwargs_single_band(self):
        """

        :return: merged kwargs from camera and obs dicts
        """
        kwargs = util.merge_dicts(self.camera, self.obs)
        return kwargs



__author__ = "aymgal, johannesulf"

import os
import shutil
import numpy as np
import sys
import copy

from lenstronomy.Sampling.Samplers.base_nested_sampler import NestedSampler

__all__ = ["DyPolyChordSampler"]


class DyPolyChordSampler(NestedSampler):
    """Wrapper for dynamical nested sampling algorithm DyPolyChord by E. Higson, M.
    Hobson, W. Handley, A. Lasenby.

    papers : arXiv:1704.03459, arXiv:1804.06406
    doc : https://dypolychord.readthedocs.io
    """

    def __init__(
        self,
        likelihood_module,
        prior_type="uniform",
        prior_means=None,
        prior_sigmas=None,
        width_scale=1,
        sigma_scale=1,
        output_dir=None,
        output_basename="-",
        resume_dyn_run=False,
        polychord_settings=None,
        remove_output_dir=False,
        use_mpi=False,
    ):  # , num_mpi_procs=1):
        """
        :param likelihood_module: likelihood_module like in likelihood.py (should be callable)
        :param prior_type: 'uniform' of 'gaussian', for converting the unit hypercube to param cube
        :param prior_means: if prior_type is 'gaussian', mean for each param
        :param prior_sigmas: if prior_type is 'gaussian', std dev for each param
        :param width_scale: scale the widths of the parameters space by this factor
        :param sigma_scale: if prior_type is 'gaussian', scale the gaussian sigma by this factor
        :param output_dir: name of the folder that will contain output files
        :param output_basename: prefix for output files
        :param resume_dyn_run: if True, previous resume files will not be deleted so that previous run can be resumed
        :param polychord_settings: settings dictionary to send to pypolychord. Check dypolychord documentation for details.
        :param remove_output_dir: remove the output_dir folder after completion
        :param use_mpi: Use MPI computing if `True`
        """
        self._check_install()
        super(DyPolyChordSampler, self).__init__(
            likelihood_module,
            prior_type,
            prior_means,
            prior_sigmas,
            width_scale,
            sigma_scale,
        )

        # if use_mpi:
        #     mpi_str = 'mpirun -np {}'.format(num_mpi_procs)
        # else:
        #     mpi_str = None
        if polychord_settings == None:
            polychord_settings = {}
        self._use_mpi = use_mpi

        self._output_dir = output_dir
        self._is_master = True

        if self._use_mpi:
            from mpi4py import MPI

            self._comm = MPI.COMM_WORLD

            if self._comm.Get_rank() != 0:
                self._is_master = False
        else:
            self._comm = None

        if not resume_dyn_run:
            if self._is_master:
                if os.path.exists(self._output_dir):
                    shutil.rmtree(self._output_dir, ignore_errors=True)
                os.mkdir(self._output_dir)

        self._output_basename = output_basename
        self._settings = copy.deepcopy(polychord_settings)
        self._settings["file_root"] = self._output_basename
        self._settings["base_dir"] = self._output_dir

        if self._all_installed:
            # create the dyPolyChord callable object
            self._sampler = self._RunPyPolyChord(
                self.log_likelihood, self.prior, self.n_dims
            )
        else:
            self._sampler = None

        self._rm_output = remove_output_dir
        self._has_warned = False

    def run(self, dynamic_goal, kwargs_run):
        """Run the DyPolyChord dynamical nested sampler.

        see https://dypolychord.readthedocs.io for content of kwargs_run

        :param dynamic_goal: 0 for evidence computation, 1 for posterior computation
        :param kwargs_run: kwargs directly passed to dyPolyChord.run_dypolychord
        :return: samples, means, logZ, logZ_err, logL, ns_run
        """
        print("prior type :", self.prior_type)
        print("parameter names :", self.param_names)

        if self._all_installed:
            # TODO : put a default dynamic_goal ?
            # dynamic_goal = 0 for evidence-only, 1 for posterior-only

            self._dyPolyChord.run_dypolychord(
                self._sampler,
                dynamic_goal,
                settings_dict_in=self._settings,
                comm=self._comm,
                **kwargs_run
            )

            if self._is_master:
                ns_run = self._ns_process_run(
                    self._settings["file_root"], self._settings["base_dir"]
                )

        else:
            # in case DyPolyChord or NestCheck was not compiled properly, for unit tests
            ns_run = {
                "theta": np.zeros((1, self.n_dims)),
                "logl": np.zeros(1),
                "output": {
                    "logZ": np.zeros(1),
                    "logZerr": np.zeros(1),
                    "param_means": np.zeros(self.n_dims),
                },
            }
            self._write_equal_weights(ns_run["theta"], ns_run["logl"])

        if self._is_master:
            samples, logL = self._get_equal_weight_samples()
            # logL     = ns_run['logl']
            # samples_w = ns_run['theta']
            logZ = ns_run["output"]["logZ"]
            logZ_err = ns_run["output"]["logZerr"]
            means = ns_run["output"]["param_means"]

            print("The log evidence estimate using the first run is {}".format(logZ))
            print("The estimated mean of the first parameter is {}".format(means[0]))

            if self._rm_output:
                shutil.rmtree(self._output_dir, ignore_errors=True)

            return samples, means, logZ, logZ_err, logL, ns_run
        else:
            sys.exit(0)

    def log_likelihood(self, args):
        """Compute the log-likelihood given list of parameters.

        :param args: parameter values
        :return: log-likelihood (from the likelihood module)
        """
        return super().log_likelihood(args), []

    def _get_equal_weight_samples(self):
        """Inspired by pymultinest's Analyzer, because DyPolyChord has more or less the
        same output conventions as MultiNest."""
        file_name = "{}_equal_weights.txt".format(self._output_basename)
        file_path = os.path.join(self._output_dir, file_name)
        data = np.loadtxt(file_path, ndmin=2)
        logL = -0.5 * data[:, 0]
        samples = data[:, 1:]
        return samples, logL

    def _write_equal_weights(self, samples, logL):
        # write fake output file for unit tests
        file_name = "{}_equal_weights.txt".format(self._output_basename)
        file_path = os.path.join(self._output_dir, file_name)
        data = np.zeros((samples.shape[0], 1 + samples.shape[1]), dtype=float)
        data[:, 0] = -2.0 * logL
        data[:, 1:] = samples
        np.savetxt(file_path, data, fmt="% .14E")

    def _check_install(self):
        try:
            import dyPolyChord
            from dyPolyChord import pypolychord_utils
        except ImportError:
            dyPolyChord = None
            pypolychord_utils = None
            print(
                "Warning : dyPolyChord not properly installed. \
                   You can get it from : https://github.com/ejhigson/dyPolyChord"
            )
            dypolychord_installed = False
        else:
            dypolychord_installed = True
            self._dyPolyChord = dyPolyChord
            self._RunPyPolyChord = pypolychord_utils.RunPyPolyChord

        try:
            from nestcheck import data_processing
        except ImportError:
            print(
                "Warning : nestcheck not properly installed (results might be unexpected). \
                   You can get it from : https://github.com/ejhigson/nestcheck"
            )
            nestcheck_installed = False
        else:
            nestcheck_installed = True
            self._ns_process_run = data_processing.process_polychord_run

        self._all_installed = dypolychord_installed and nestcheck_installed



import numpy as np


def auto_raytracing_grid_size(source_fwhm_parcsec, grid_size_scale=0.005, power=1.0):
    """This function returns the size of a ray tracing grid in units of arcsec
    appropriate for magnification computations with finite-size background sources. This
    fit is calibrated for source sizes (interpreted as the FWHM of a Gaussian) in the
    range 0.1 -100 pc.

    :param source_fwhm_parcsec: the full width at half max of a Gaussian background
        source
    :return: an appropriate grid size for finite-size background magnification
        computation
    """

    grid_radius_arcsec = grid_size_scale * source_fwhm_parcsec**power
    return grid_radius_arcsec


def auto_raytracing_grid_resolution(
    source_fwhm_parcsec, grid_resolution_scale=0.0002, ref=10.0, power=1.0
):
    """This function returns a resolution factor in units arcsec/pixel appropriate for
    magnification computations with finite-size background sources. This fit is
    calibrated for source sizes (interpreted as the FWHM of a Gaussian) in the range
    0.1-100 pc.

    :param source_fwhm_parcsec: the full width at half max of a Gaussian background
        source
    :return: an appropriate grid resolution for finite-size background magnification
        computation
    """

    grid_resolution = grid_resolution_scale * (source_fwhm_parcsec / ref) ** power
    return grid_resolution


def setup_mag_finite(grid_radius_arcsec, grid_resolution, source_model, kwargs_source):
    """Sets up the ray tracing grid and source light model for
    magnification_finite_adaptive and plot_quasar_images routines.

    This new updates allows for more flexibility in the source light model by requiring
    the user to specify the source light mode, grid size and grid resolution before
    calling the function.

    The functions auto_raytrracing_grid_size and auto_raytracing_grid_resolution give
    good estimates for appropriate parameter choices for grid_radius_arcsec and
    grid_resolution.

    :param grid_radius_arcsec: the size of the ray tracing region in arcsec; if not
        specified, an appropriate value will be estimated from the source size
    :param grid_resolution: the grid resolution in units arcsec/pixel; if not specified,
        an appropriate value will be estimated from the source size
    :param source_model: instance of LightModel for the source :kwargs_source: keyword
        arguments for the light profile corresponding to the desired LightModel instance
    """

    # setup the grid
    npix = int(2 * grid_radius_arcsec / grid_resolution)
    _grid_x = np.linspace(-grid_radius_arcsec, grid_radius_arcsec, npix)
    _grid_y = np.linspace(-grid_radius_arcsec, grid_radius_arcsec, npix)
    grid_x_0, grid_y_0 = np.meshgrid(_grid_x, _grid_y)

    return (
        grid_x_0,
        grid_y_0,
        source_model,
        kwargs_source,
        grid_resolution,
        grid_radius_arcsec,
    )



from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
import numpy as np

__all__ = ["FluxRatioLikelihood"]


class FluxRatioLikelihood(object):
    """Likelihood class for magnification of multiply lensed images."""

    def __init__(
        self,
        lens_model_class,
        flux_ratios,
        flux_ratio_errors,
        flux_ratio_measurement_bool=None,
        num_point_sources=1,
        source_type="INF",
        window_size=0.1,
        grid_number=100,
        polar_grid=False,
        aspect_ratio=0.5,
        point_source_redshift_list=None,
    ):
        """

        :param lens_model_class: LensModel class instance
        :param flux_ratios: ratio of fluxes of the multiple images (relative to the first appearing in
         same order as the images)
        :param flux_ratio_errors: errors in the flux ratios (relative to the first appearing). Alternatively
         a log-normal covariance matrix. Note: in the case of a covariance matrix, the errors are assumed
         to be log-normal, i.e. the logarithms of the flux ratios, ln(F[i]/F[0]) are assumed to have a multivariate
         Gaussian distribution, with the given covariance matrix.
        :param flux_ratio_measurement_bool: list of bools of length of the point source model,
         indicating for which point sources there is a flux ratio measurement
        :param num_point_sources: number of point sources in the point source model output
        :type num_point_sources: int>=0
        :param source_type: string, type of source, 'INF' specifies a point source, while 'GAUSSIAN' specifies a
         finite-size source modeled as a Gaussian
        :param window_size: size of window to compute the finite flux
        :param grid_number: number of grid cells per axis in the window to numerically compute the flux
        :param point_source_redshift_list: list of redshifts to the different sources
        """
        self._num_point_sources = num_point_sources
        self._lens_model_class = lens_model_class
        if num_point_sources == 1:
            self._flux_ratios = [np.array(flux_ratios)]
            self._flux_ratio_errors = [np.array(flux_ratio_errors)]
        else:
            self._flux_ratios = []
            self._flux_ratio_errors = []
            for i in range(num_point_sources):
                self._flux_ratios.append(np.array(flux_ratios[i]))
                self._flux_ratio_errors.append(np.array(flux_ratio_errors[i]))
        self._lens_model_extensions = LensModelExtensions(lensModel=lens_model_class)
        self._source_type = source_type
        self._window_size = window_size
        self._gird_number = grid_number
        self._polar_grid = polar_grid
        self._aspect_ratio = aspect_ratio
        self._num_point_sources = num_point_sources
        if flux_ratio_measurement_bool is None:
            flux_ratio_measurement_bool = [True] * num_point_sources
        self._flux_ratio_measurement_bool = flux_ratio_measurement_bool
        if point_source_redshift_list is None:
            point_source_redshift_list = [None] * num_point_sources
        self._point_source_redshift_list = point_source_redshift_list

    def logL(self, ra_image_list, dec_image_list, kwargs_lens, kwargs_special):
        """

        :param kwargs_lens: lens model keyword argument list
        :param kwargs_special: dictionary of 'special' keyword parameters
        :return: log likelihood of the measured flux ratios given a model
        """
        log_l = 0
        for k in range(len(ra_image_list)):
            if self._flux_ratio_measurement_bool[k] is True:
                x_pos, y_pos = ra_image_list[k], dec_image_list[k]
                self._lens_model_class.change_source_redshift(
                    z_source=self._point_source_redshift_list[k]
                )
                if self._source_type == "INF":
                    mag = np.abs(
                        self._lens_model_class.magnification(x_pos, y_pos, kwargs_lens)
                    )
                else:
                    source_sigma = kwargs_special["source_size"]
                    mag = self._lens_model_extensions.magnification_finite(
                        x_pos,
                        y_pos,
                        kwargs_lens,
                        source_sigma=source_sigma,
                        window_size=self._window_size,
                        grid_number=self._gird_number,
                        polar_grid=self._polar_grid,
                        aspect_ratio=self._aspect_ratio,
                    )
                if len(mag) - 1 != len(self._flux_ratios[k]):
                    return -(10**15)
                mag_ratio = mag[1:] / mag[0]
                log_l += self._logL(mag_ratio, k=k)
        return log_l

    def _logL(self, flux_ratios, k=0):
        """


        :param flux_ratios: flux ratios from the model
        :param k: integer listing the n'th point source model
        :return: log likelihood fo the measured flux ratios given a model
        """
        flux_ratios_obs = self._flux_ratios[k]
        flux_ratio_errors = self._flux_ratio_errors[k]
        if not np.isfinite(flux_ratios).any():
            return -(10**15)
        if flux_ratio_errors.ndim <= 1:
            dist = (flux_ratios - flux_ratios_obs) ** 2 / flux_ratio_errors**2 / 2
            logL = -np.sum(dist)
        elif flux_ratio_errors.ndim == 2:
            # Assume covariance matrix is in ln units!
            D = np.log(flux_ratios) - np.log(flux_ratios_obs)
            logL = (
                -1 / 2 * D @ np.linalg.inv(flux_ratio_errors) @ D
            )  # TODO: only calculate the inverse once
        else:
            raise ValueError(
                "flux_ratio_errors need dimension of 1 or 2. Current dimensions are %s"
                % self._flux_ratio_errors.ndim
            )
        if not np.isfinite(logL):
            return -(10**15)
        return logL

    @property
    def num_data(self):
        """

        :return: integer, number of data points associated with the flux ratios
        """
        num = 0
        for k in range(self._num_point_sources):
            if self._flux_ratio_measurement_bool[k] is True:
                num += len(self._flux_ratios[k])
        return num



from lenstronomy.Conf import config_loader
from os import environ

"""
From pyautolens:
Depending on if we're using a super computer, we want two different numba decorators:
If on laptop:
@numba.jit(nopython=True, cache=True, parallel=False)
If on super computer:
@numba.jit(nopython=True, cache=False, parallel=True)
"""

numba_conf = config_loader.numba_conf()
nopython = numba_conf["nopython"]
cache = numba_conf["cache"]
parallel = numba_conf["parallel"]
numba_enabled = numba_conf["enable"] and not environ.get("NUMBA_DISABLE_JIT", False)
fastmath = numba_conf["fastmath"]
error_model = numba_conf["error_model"]

if numba_enabled:
    try:
        import numba
        from numba import extending
    except ImportError:
        numba_enabled = False
        numba = None
        extending = None

__all__ = ["jit"]


def jit(
    nopython=nopython,
    cache=cache,
    parallel=parallel,
    fastmath=fastmath,
    error_model=error_model,
    inline="never",
):
    if numba_enabled:

        def wrapper(func):
            return numba.jit(
                func,
                nopython=nopython,
                cache=cache,
                parallel=parallel,
                fastmath=fastmath,
                error_model=error_model,
                inline=inline,
            )

    else:

        def wrapper(func):
            return func

    return wrapper



__author__ = "lynevdv"

import numpy as np

import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["Multipole", "EllipticalMultipole"]


class Multipole(LensProfileBase):
    """
    This class contains a SPHERICAL multipole contribution (for 1 component with m>=2)
    This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf, Equation B12
    Only the q=1 case (ie., spherical symmetry) makes this definition consistent with interpretation of multipoles as a deformation of the isophotes with an order m symmetry (eg., disky/boxy in the m=4 case).

    m : int, multipole order, m>=2
    a_m : float, multipole strength
    phi_m : float, multipole orientation in radian
    """

    param_names = ["m", "a_m", "phi_m", "center_x", "center_y"]
    lower_limit_default = {
        "m": 2,
        "a_m": 0,
        "phi_m": -np.pi,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "m": 100,
        "a_m": 100,
        "phi_m": np.pi,
        "center_x": 100,
        "center_y": 100,
    }

    def function(self, x, y, m, a_m, phi_m, center_x=0, center_y=0):
        """
        Lensing potential of multipole contribution (for 1 component with m>=2)
        This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf

        :param x: x-coordinate to evaluate function
        :param y: y-coordinate to evaluate function
        :param m: int, multipole order, m>=2
        :param a_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: y-position
        :return: lensing potential
        """

        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        f_ = r * a_m / (1 - m**2) * np.cos(m * (phi - phi_m))
        return f_

    def derivatives(self, x, y, m, a_m, phi_m, center_x=0, center_y=0):
        """
        Deflection of a multipole contribution (for 1 component with m>=2)
        This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf
        Equation B12

        :param x: x-coordinate to evaluate function
        :param y: y-coordinate to evaluate function
        :param m: int, multipole order, m>=2
        :param a_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: y-position
        :return: deflection angles alpha_x, alpha_y
        """

        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        f_x = np.cos(phi) * a_m / (1 - m**2) * np.cos(m * (phi - phi_m)) + np.sin(
            phi
        ) * m * a_m / (1 - m**2) * np.sin(m * (phi - phi_m))
        f_y = np.sin(phi) * a_m / (1 - m**2) * np.cos(m * (phi - phi_m)) - np.cos(
            phi
        ) * m * a_m / (1 - m**2) * np.sin(m * (phi - phi_m))
        return f_x, f_y

    def hessian(self, x, y, m, a_m, phi_m, center_x=0, center_y=0):
        """
        Hessian of a multipole contribution (for 1 component with m>=2)
        This uses the same definitions as Xu et al.(2013) in Appendix B3 https://arxiv.org/pdf/1307.4220.pdf

        :param x: x-coordinate to evaluate function
        :param y: y-coordinate to evaluate function
        :param m: int, multipole order, m>=2
        :param a_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: y-position
        :return: f_xx, f_xy, f_yx, f_yy
        """

        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        r = np.maximum(r, 0.000001)
        f_xx = 1.0 / r * np.sin(phi) ** 2 * a_m * np.cos(m * (phi - phi_m))
        f_yy = 1.0 / r * np.cos(phi) ** 2 * a_m * np.cos(m * (phi - phi_m))
        f_xy = -1.0 / r * a_m * np.cos(phi) * np.sin(phi) * np.cos(m * (phi - phi_m))
        return f_xx, f_xy, f_xy, f_yy


class EllipticalMultipole(LensProfileBase):
    """This class contains a multipole contribution that encode deviations from the
    elliptical isodensity contours of a SIE with any axis ratio q.

    m : int, multipole order, m=3 or m=4
    a_m : float, multipole strength
    phi_m : float, multipole orientation in radian
    q : axis ratio of the reference ellipses
    """

    param_names = ["m", "a_m", "phi_m", "q", "center_x", "center_y"]
    lower_limit_default = {
        "m": 2,
        "a_m": 0,
        "phi_m": -np.pi,
        "q": 0.001,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "m": 100,
        "a_m": 100,
        "phi_m": np.pi,
        "q": 1,
        "center_x": 100,
        "center_y": 100,
    }

    def function(self, x, y, m, a_m, phi_m, q, center_x=0, center_y=0):
        """Lensing potential of multipole contribution (for 1 component with m=3 or m=4)

        :param x: x-coordinate to evaluate function
        :param y: y-coordinate to evaluate function
        :param m: int, multipole order, m>=2
        :param a_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: y-position
        :return: lensing potential
        """

        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)

        if np.abs(1 - q) < 1e-4:  # avoid numerical instability when q is too close to 1
            f_ = r * a_m / (1 - m**2) * np.cos(m * (phi - phi_m))

        else:
            if m == 3:
                ### Symmetrize at this level to avoid discontinuities
                phisymm = param_util.cart2polar(-r * np.cos(phi), r * np.sin(phi))[1]
                phirot = (
                    param_util.cart2polar(-r * np.sin(phi), r * np.cos(phi))[1]
                    - np.pi / 2
                )  # find angle corresponding to phi in (-3pi/2, pi/2]
                phirotsymm = (
                    param_util.cart2polar(r * np.sin(phi), r * np.cos(phi))[1]
                    - np.pi / 2
                )  # find angle corresponding to -phi in (-3pi/2, pi/2]
                f_ = (
                    a_m
                    * np.sqrt(q)
                    * r
                    * (
                        (_F_m3_1(phi, q) - _F_m3_1(phisymm, q)) / 2 * np.cos(m * phi_m)
                        + (_F_m3_2(phirot, q) - _F_m3_2(phirotsymm, q))
                        / 2
                        * np.sin(m * phi_m)
                    )
                )

            elif m == 4:
                f_ = (
                    a_m
                    * np.sqrt(q)
                    * r
                    * (
                        _F_m4_1(phi, q=q) * np.cos(m * phi_m)
                        + _F_m4_2(phi, q=q) * np.sin(m * phi_m)
                    )
                )

            else:
                raise ValueError(
                    "Implementation of multipoles perturbation with m>4 for general axis ratio q not available."
                )

        return f_

    def derivatives(self, x, y, m, a_m, phi_m, q, center_x=0, center_y=0):
        """Deflection of a multipole contribution (for 1 component with m=3 or m=4)

        :param x: x-coordinate to evaluate function
        :param y: y-coordinate to evaluate function
        :param m: int, multipole order, m>=2
        :param a_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: y-position
        :return: deflection angles alpha_x, alpha_y
        """

        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)

        if np.abs(1 - q) < 1e-4:  # avoid numerical instability when q is too close to 1
            f_x = np.cos(phi) * a_m / (1 - m**2) * np.cos(m * (phi - phi_m)) + np.sin(
                phi
            ) * m * a_m / (1 - m**2) * np.sin(m * (phi - phi_m))
            f_y = np.sin(phi) * a_m / (1 - m**2) * np.cos(m * (phi - phi_m)) - np.cos(
                phi
            ) * m * a_m / (1 - m**2) * np.sin(m * (phi - phi_m))

        else:
            if m == 3:
                ### Symmetrize at this level to avoid discontinuities
                phisymm = param_util.cart2polar(-r * np.cos(phi), r * np.sin(phi))[1]
                phirot = (
                    param_util.cart2polar(-r * np.sin(phi), r * np.cos(phi))[1]
                    - np.pi / 2
                )  # find angle corresponding to phi in (-3pi/2, pi/2]
                phirotsymm = (
                    param_util.cart2polar(r * np.sin(phi), r * np.cos(phi))[1]
                    - np.pi / 2
                )  # find angle corresponding to -phi in (-3pi/2, pi/2]
                alpha_x_1_pos, alpha_y_1_pos = _deflection_m3_base(
                    r, phi, q, a_3=a_m, phi_3=0
                )
                alpha_x_1_neg, alpha_y_1_neg = _deflection_m3_base(
                    r, phisymm, q, a_3=a_m, phi_3=0
                )
                alpha_x_2_pos, alpha_y_2_pos = _deflection_m3_base(
                    r, phirot, q, a_3=a_m, phi_3=np.pi / 6
                )
                alpha_x_2_neg, alpha_y_2_neg = _deflection_m3_base(
                    r, phirotsymm, q, a_3=a_m, phi_3=np.pi / 6
                )
                f_x = (
                    np.cos(m * phi_m) * (alpha_x_1_pos + alpha_x_1_neg) / 2
                    + np.sin(m * phi_m) * (alpha_x_2_pos - alpha_x_2_neg) / 2
                )
                f_y = (
                    np.cos(m * phi_m) * (alpha_y_1_pos - alpha_y_1_neg) / 2
                    + np.sin(m * phi_m) * (alpha_y_2_pos + alpha_y_2_neg) / 2
                )

            elif m == 4:
                F_m4 = _F_m4_1(phi, q=q) * np.cos(m * phi_m) + _F_m4_2(
                    phi, q=q
                ) * np.sin(m * phi_m)
                F_m4_prime = _F_m4_1_derivative(phi, q=q) * np.cos(
                    m * phi_m
                ) + _F_m4_2_derivative(phi, q=q) * np.sin(m * phi_m)
                f_x = a_m * np.sqrt(q) * (F_m4 * np.cos(phi) - F_m4_prime * np.sin(phi))
                f_y = a_m * np.sqrt(q) * (F_m4 * np.sin(phi) + F_m4_prime * np.cos(phi))

            else:
                raise ValueError(
                    "Implementation of multipoles perturbation with m>4 for general axis ratio q not available."
                )

        return f_x, f_y

    def hessian(self, x, y, m, a_m, phi_m, q, center_x=0, center_y=0):
        """Hessian of a multipole contribution (for 1 component with m=3 or m=4)

        :param x: x-coordinate to evaluate function
        :param y: y-coordinate to evaluate function
        :param m: int, multipole order, m>=2
        :param a_m: float, multipole strength
        :param phi_m: float, multipole orientation in radian
        :param center_x: x-position
        :param center_y: y-position
        :return: f_xx, f_xy, f_yx, f_yy
        """
        r, phi = param_util.cart2polar(x, y, center_x=center_x, center_y=center_y)
        r = np.maximum(r, 0.000001)
        phi_ell = np.angle(q * r * np.cos(phi) + 1j * r * np.sin(phi))
        R = np.sqrt(q * (r * np.cos(phi)) ** 2 + (r * np.sin(phi)) ** 2 / q)

        delta_r = a_m * np.cos(m * (phi_ell - phi_m)) * r / R
        f_xx = np.sin(phi) ** 2 * delta_r / r
        f_yy = np.cos(phi) ** 2 * delta_r / r
        f_xy = -np.sin(phi) * np.cos(phi) * delta_r / r
        return f_xx, f_xy, f_xy, f_yy


def _phi_ell(phi, q):
    return (
        phi
        - np.arctan2(np.sin(phi), np.cos(phi))
        + np.arctan2(np.sin(phi), q * np.cos(phi))
    )


def _F_m3_1(phi, q):
    term1 = np.cos(phi) * (
        q * (3 + q**2) * np.log(1 + q**2 + (q**2 - 1) * np.cos(2 * phi))
        - (
            np.log(2) * (1 + q) ** 2
            - 2 * (1 - q) * (1 + q) ** 2 * (1 + np.log(2) / 4)
            + (1 - q**2) ** 2 / 4
        )
    )
    term2 = (
        2 * np.sin(phi) * (q * (q**2 + 3) * phi - (1 + 3 * q**2) * _phi_ell(phi, q))
    )  # Expression valid in (-pi, pi]
    return (term1 + term2) / (2 * (1 - q**2) ** 2)


def _F_m3_1_derivative(phi, q):
    term1 = -np.cos(phi) * q * (3 + q**2) * 2 * (q**2 - 1) * np.sin(2 * phi) / (
        1 + q**2 + (q**2 - 1) * np.cos(2 * phi)
    ) + np.sin(phi) * (
        -q * (3 + q**2) * np.log(1 + q**2 + (q**2 - 1) * np.cos(2 * phi))
        + np.log(2) * (1 + q) ** 2
        - 2 * (1 - q) * (1 + q) ** 2 * (1 + np.log(2) / 4)
        + (1 - q**2) ** 2 / 4
    )
    term2 = 2 * np.cos(phi) * (
        q * (q**2 + 3) * phi - (1 + 3 * q**2) * _phi_ell(phi, q)
    ) + 2 * np.sin(phi) * (
        q * (q**2 + 3)
        - (1 + 3 * q**2) * q / (q**2 * np.cos(phi) ** 2 + np.sin(phi) ** 2)
    )  # Expression valid in (-pi, pi]
    return (term1 + term2) / (2 * (1 - q**2) ** 2)


def _F_m3_2(phi, q):
    # Expression valid in (-3*pi/2, pi/2]
    return 1 / q * _F_m3_1(phi + np.pi / 2, 1 / q)


def _F_m3_2_derivative(phi, q):
    # Expression valid in (-3*pi/2, pi/2]
    return 1 / q * _F_m3_1_derivative(phi + np.pi / 2, 1 / q)


def _deflection_m3_base(r, phi, q, a_3, phi_3):
    F_m3 = _F_m3_1(phi, q=q) * np.cos(3 * phi_3) + _F_m3_2(phi, q=q) * np.sin(3 * phi_3)
    F_m3_prime = _F_m3_1_derivative(phi, q=q) * np.cos(3 * phi_3) + _F_m3_2_derivative(
        phi, q=q
    ) * np.sin(3 * phi_3)
    alpha_x = a_3 * np.sqrt(q) * (F_m3 * np.cos(phi) - F_m3_prime * np.sin(phi))
    alpha_y = a_3 * np.sqrt(q) * (F_m3 * np.sin(phi) + F_m3_prime * np.cos(phi))
    return alpha_x, alpha_y


def _F_m4_1(phi, q):
    term1 = (
        -4
        * np.sqrt(2)
        * (1 + 4 * q**2 + q**4 + (q**4 - 1) * np.cos(2 * phi))
        / ((3 * (1 - q**2) ** 2) * np.sqrt(1 + q**2 + (q**2 - 1) * np.cos(2 * phi)))
    )
    term2 = (
        (1 + 6 * q**2 + q**4)
        / (1 - q**2) ** (5 / 2)
        * np.cos(phi)
        * np.arctan(
            (np.sqrt(2 * (1 - q**2)) * np.cos(phi))
            / np.sqrt(1 + q**2 + (q**2 - 1) * np.cos(2 * phi))
        )
    )
    term3 = (
        (1 + 6 * q**2 + q**4)
        / (1 - q**2) ** (5 / 2)
        * np.sin(phi)
        * np.log(
            np.sqrt(1 - q**2) * np.sin(phi) / q
            + np.sqrt(1 + (1 - q**2) / q**2 * np.sin(phi) ** 2)
        )
    )

    return term1 + term2 + term3


def _F_m4_1_derivative(phi, q):
    term1 = (
        -4
        * np.sqrt(2)
        * (1 + q**4 + (q**4 - 1) * np.cos(2 * phi))
        * np.sin(2 * phi)
        / (3 * (1 - q**2) * (1 + q**2 + (q**2 - 1) * np.cos(2 * phi)) ** (3 / 2))
    )
    term2 = (
        -(1 + 6 * q**2 + q**4)
        / (1 - q**2) ** (5 / 2)
        * (
            np.sin(phi)
            * np.arctan(
                (np.sqrt(2 * (1 - q**2)) * np.cos(phi))
                / np.sqrt(1 + q**2 + (q**2 - 1) * np.cos(2 * phi))
            )
            + np.sqrt(2 * (1 - q**2))
            * np.sin(2 * phi)
            / (2 * np.sqrt(1 + q**2 + (q**2 - 1) * np.cos(2 * phi)))
        )
    )
    term3 = (
        (1 + 6 * q**2 + q**4)
        / (1 - q**2) ** (5 / 2)
        * np.cos(phi)
        * (
            np.log(
                np.sqrt(1 - q**2) * np.sin(phi) / q
                + np.sqrt(1 + (1 - q**2) / q**2 * np.sin(phi) ** 2)
            )
            + np.sqrt(1 - q**2)
            / q
            * np.sin(phi)
            / np.sqrt(1 + (1 - q**2) / q**2 * np.sin(phi) ** 2)
        )
    )

    return term1 + term2 + term3


def _F_m4_2(phi, q):
    term1 = (
        -4
        * np.sqrt(2)
        * q
        / (3 * (1 - q**2))
        * np.sin(2 * phi)
        / np.sqrt(1 + q**2 + (q**2 - 1) * np.cos(2 * phi))
    )
    term2 = (
        -4
        * q
        * (1 + q**2)
        / (1 - q**2) ** (5 / 2)
        * np.sin(phi)
        * np.arctan(
            (np.sqrt(2 * (1 - q**2)) * np.cos(phi))
            / np.sqrt(1 + q**2 + (q**2 - 1) * np.cos(2 * phi))
        )
    )
    term3 = (
        4
        * q
        * (1 + q**2)
        / (1 - q**2) ** (5 / 2)
        * np.cos(phi)
        * np.log(
            np.sqrt(1 - q**2) * np.sin(phi) / q
            + np.sqrt(1 + (1 - q**2) / q**2 * np.sin(phi) ** 2)
        )
    )

    return term1 + term2 + term3


def _F_m4_2_derivative(phi, q):
    term1 = (
        -8
        * np.sqrt(2)
        * q
        / (6 * (1 - q**2))
        * (
            -(1 - q**2)
            * np.sin(2 * phi) ** 2
            / (1 + q**2 + (q**2 - 1) * np.cos(2 * phi)) ** (3 / 2)
            + 2 * np.cos(2 * phi) / np.sqrt(1 + q**2 + (q**2 - 1) * np.cos(2 * phi))
        )
    )
    term2 = (
        4
        * q
        * (1 + q**2)
        / (1 - q**2) ** (5 / 2)
        * (
            -np.cos(phi)
            * np.arctan(
                (np.sqrt(2 * (1 - q**2)) * np.cos(phi))
                / np.sqrt(1 + q**2 + (q**2 - 1) * np.cos(2 * phi))
            )
            + 2
            * np.sqrt(2 * (1 - q**2))
            * np.sin(phi) ** 2
            / (2 * np.sqrt(1 + q**2 + (q**2 - 1) * np.cos(2 * phi)))
        )
    )
    term3 = (
        4
        * q
        * (1 + q**2)
        / (1 - q**2) ** (5 / 2)
        * (
            -np.sin(phi)
            * np.log(
                np.sqrt(1 - q**2) * np.sin(phi) / q
                + np.sqrt(1 + (1 - q**2) / q**2 * np.sin(phi) ** 2)
            )
            + np.sqrt(1 - q**2)
            / q
            * np.cos(phi) ** 2
            / np.sqrt(1 + (1 - q**2) / q**2 * np.sin(phi) ** 2)
        )
    )

    return term1 + term2 + term3



from lenstronomy.Util.magnification_finite_util import setup_mag_finite
import numpy as np
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions


def plot_quasar_images(
    lens_model,
    x_image,
    y_image,
    kwargs_lens,
    source_light_model,
    kwargs_light_source,
    grid_resolution,
    grid_radius_arcsec,
):
    """This function plots the surface brightness in the image plane of a background
    source. The flux is computed inside a circular aperture with radius
    grid_radius_arcsec. This new updates allows for more flexibility in the source light
    model by requiring the user to specify the source light mode, grid size and grid
    resolution before calling the function. The functions auto_raytrracing_grid_size and
    auto_raytracing_grid_resolution give good estimates for appropriate parameter
    choices for grid_radius_arcsec and grid_resolution.

    :param lens_model: an instance of LensModel
    :param x_image: a list or array of x coordinates [units arcsec]
    :param y_image: a list or array of y coordinates [units arcsec]
    :param kwargs_lens: keyword arguments for the lens model
    :param source_size: the size of the background source [units parsec]
    :param source_light_model: instance of LightModel for the source
    :param source_light_kwargs: the keyword arguments for the source light
    :param grid_resolution: the grid resolution in units arcsec/pixel
    :param grid_radius_arcsec: the size of the ray tracing region in arcsec
    """
    magnifications = []
    images = []

    (
        grid_x_0,
        grid_y_0,
        source_model,
        kwargs_source,
        grid_resolution,
        grid_radius_arcsec,
    ) = setup_mag_finite(
        grid_radius_arcsec, grid_resolution, source_light_model, kwargs_light_source
    )
    shape0 = grid_x_0.shape
    grid_x_0, grid_y_0 = grid_x_0.ravel(), grid_y_0.ravel()
    for k, (xi, yi) in enumerate(zip(x_image, y_image)):
        flux_array = np.zeros_like(grid_x_0)
        r_min = 0
        r_max = grid_radius_arcsec
        grid_r = np.hypot(grid_x_0, grid_y_0)
        lens_model_extension = LensModelExtensions(lens_model)
        flux_array = lens_model_extension._magnification_adaptive_iteration(
            flux_array,
            xi,
            yi,
            grid_x_0,
            grid_y_0,
            grid_r,
            r_min,
            r_max,
            lens_model,
            kwargs_lens,
            source_model,
            kwargs_source,
        )
        m = np.sum(flux_array) * grid_resolution**2
        magnifications.append(m)
        images.append(flux_array.reshape(shape0))

    magnifications = np.array(magnifications)
    flux_ratios = magnifications / max(magnifications)
    import matplotlib.pyplot as plt

    N = len(images)
    fig, axes = plt.subplots(N, 1, figsize=(16, 6))
    for i, (image, mag, fr) in enumerate(zip(images, magnifications, flux_ratios)):
        ax = axes[i]
        ax.imshow(
            image,
            origin="lower",
            extent=[
                -grid_radius_arcsec,
                grid_radius_arcsec,
                -grid_radius_arcsec,
                grid_radius_arcsec,
            ],
        )
        ax.annotate(
            "magnification: " + str(np.round(mag, 3)),
            xy=(0.05, 0.9),
            xycoords="axes fraction",
            color="w",
            fontsize=12,
        )
        ax.annotate(
            "flux ratio: " + str(np.round(fr, 3)),
            xy=(0.05, 0.8),
            xycoords="axes fraction",
            color="w",
            fontsize=12,
        )
    plt.show()



import pytest
import numpy as np
import numpy.testing as npt
from lenstronomy.LensModel.Profiles.blank_plane import BlankPlane


class TestBlankPlane(object):
    """Class to test the "BLANK_PLANE" lens model."""

    def setup_method(self):
        self.blank = BlankPlane()

    def test_function(self):
        """

        :return:
        """
        x = np.linspace(0.1, 10, 10)
        f_ = self.blank.function(x, x)
        npt.assert_almost_equal(f_, x * 0, decimal=5)

    def test_derivatives(self):
        """

        :return:
        """
        x = np.linspace(0.1, 10, 10)
        f_x_1, f_y_1 = self.blank.derivatives(x, x)
        npt.assert_almost_equal(f_x_1, x * 0, decimal=5)
        npt.assert_almost_equal(f_y_1, x * 0, decimal=5)

    def test_hessian(self):
        """

        :return:
        """
        x = np.linspace(0.1, 10, 10)
        f_xx, f_xy, f_yx, f_yy = self.blank.hessian(x, x)
        npt.assert_almost_equal(f_xx, x * 0, decimal=5)
        npt.assert_almost_equal(f_xy, x * 0, decimal=5)
        npt.assert_almost_equal(f_yx, x * 0, decimal=5)
        npt.assert_almost_equal(f_yy, x * 0, decimal=5)


if __name__ == "__main__":
    pytest.main()



import numpy as np
from scipy.interpolate import interp1d

import lenstronomy.Util.constants as const
from lenstronomy.GalKin.light_profile import LightProfile
from lenstronomy.GalKin.anisotropy import Anisotropy
from lenstronomy.GalKin.cosmo import Cosmo
from lenstronomy.LensModel.single_plane import SinglePlane
import lenstronomy.GalKin.velocity_util as util

__all__ = ["NumericKinematics"]


class NumericKinematics(Anisotropy):
    def __init__(
        self,
        kwargs_model,
        kwargs_cosmo,
        interpol_grid_num=200,
        log_integration=True,
        max_integrate=100,
        min_integrate=0.0001,
        max_light_draw=None,
        lum_weight_int_method=True,
    ):
        """
        What we need:
        - max projected R to have ACCURATE I_R_sigma values
        - make sure everything outside cancels out (or is not rendered)

        :param interpol_grid_num: number of interpolation bins for integrand and interpolated functions
        :param log_integration: bool, if True, performs the numerical integral in log space distance (adviced)
            (only applies for lum_weight_int_method=True). If set to False, may lead to less accurate results
        :param max_integrate: maximum radius (in arc seconds) of the Jeans equation integral
            (assumes zero tracer particles outside this radius)
        :param max_light_draw: float; (optional) if set, draws up to this radius, else uses max_interpolate value
        :param lum_weight_int_method: bool, luminosity weighted dispersion integral to calculate LOS projected Jean's
         solution. ATTENTION: currently less accurate than 3d solution
        :param min_integrate:
        """
        mass_profile_list = kwargs_model.get("mass_profile_list")
        light_profile_list = kwargs_model.get("light_profile_list")
        anisotropy_model = kwargs_model.get("anisotropy_model")
        self._interp_grid_num = interpol_grid_num
        self._log_int = log_integration
        self._max_integrate = (
            max_integrate  # maximal integration (and interpolation) in units of arcsecs
        )
        self._min_integrate = (
            min_integrate  # min integration (and interpolation) in units of arcsecs
        )
        self._max_interpolate = max_integrate  # we chose to set the interpolation range to the integration range
        self._min_interpolate = min_integrate  # we chose to set the interpolation range to the integration range
        if max_light_draw is None:
            # make sure the actual solution for the kinematics is only computed way inside the integral
            max_light_draw = max_integrate
        self.lightProfile = LightProfile(
            light_profile_list,
            interpol_grid_num=interpol_grid_num,
            max_interpolate=max_integrate,
            min_interpolate=min_integrate,
            max_draw=max_light_draw,
        )
        Anisotropy.__init__(self, anisotropy_type=anisotropy_model)
        self.cosmo = Cosmo(**kwargs_cosmo)
        self._mass_profile = SinglePlane(mass_profile_list)
        self._lum_weight_int_method = lum_weight_int_method

    @property
    def lum_weight_int_method(self):
        """Get the luminosity weighted integration method."""
        return self._lum_weight_int_method

    @property
    def max_integrate(self):
        """Get the maximum range of integration."""
        return self._max_integrate

    @property
    def min_integrate(self):
        """Get the minimum range of integration."""
        return self._min_integrate

    def lum_weighted_vel_disp(self, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Luminosity-weighted line-of-sight velocity dispersion within a radius R.

        :param R: 2d projected radius (in angular units of arcsec)
        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. We refer to the Anisotropy() class for details on
            the parameters.
        :return: average velocity dispersion [km/s]
        """
        # bins in I_R
        r_rad = np.linspace(0.0001, R, num=50)
        I_R_sigma2_rad, I_R_rad = self._I_R_sigma2_interp(
            r_rad, kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        # azimuthal averaging
        I_R_sigma2 = np.sum(I_R_sigma2_rad * r_rad)
        I_R = np.sum(I_R_rad * r_rad)
        # return in units km/s
        sigma_s2 = I_R_sigma2 / I_R
        sigma = np.sqrt(sigma_s2) / 1000.0
        return sigma

    def sigma_s2(self, r, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Returns unweighted los velocity dispersion for a specified 3d and projected
        radius (if lum_weight_int_method=True then the 3d radius is not required and the
        function directly performs the luminosity weighted integral in projection at R)

        :param r: 3d radius (not needed for this calculation)
        :param R: 2d projected radius (in angular units of arcsec)
        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. We refer to the Anisotropy() class for details on
            the parameters.
        :return: weighted line-of-sight projected velocity dispersion at projected
            radius R with weights I
        """
        if self._lum_weight_int_method is True:
            return self.sigma_s2_project(
                R, kwargs_mass, kwargs_light, kwargs_anisotropy
            )
        else:
            return (
                self.sigma_s2_r(r, R, kwargs_mass, kwargs_light, kwargs_anisotropy),
                1,
            )

    def sigma_s2_project(self, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Returns luminosity-weighted los velocity dispersion for a specified projected
        radius R and weight.

        :param R: 2d projected radius (in angular units of arcsec)
        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. We refer to the Anisotropy() class for details on
            the parameters.
        :return: line-of-sight projected velocity dispersion at projected radius R
        """
        # nominator is numerically to a finite distance, so luminosity weighting might be off
        # this could lead to an under-prediction of the velocity dispersion
        # so we ask the function _I_R_sigma2() to also return the numerical l(r)
        # I_R_sigma2, I_R = self._I_R_sigma2_interp(R, kwargs_mass, kwargs_light, kwargs_anisotropy)
        I_R_sigma2, I_R = self._I_R_sigma2_interp(
            R, kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        # I_R = self.light_profile.light_2d(R, kwargs_light)
        return I_R_sigma2 / I_R, 1

    def sigma_s2_r(self, r, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Returns unweighted los velocity dispersion for a specified 3d radius r at
        projected radius R.

        :param r: 3d radius (not needed for this calculation)
        :param R: 2d projected radius (in angular units of arcsec)
        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. We refer to the Anisotropy() class for details on
            the parameters.
        :return: line-of-sight projected velocity dispersion at projected radius R from
            3d radius r
        """
        beta = self.beta_r(r, **kwargs_anisotropy)
        return (1 - beta * R**2 / r**2) * self.sigma_r2(
            r, kwargs_mass, kwargs_light, kwargs_anisotropy
        )

    def sigma_r2(self, r, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """
        computes numerically the solution of the Jeans equation for a specific 3d radius
        E.g. Equation (A1) of Mamon & Lokas https://arxiv.org/pdf/astro-ph/0405491.pdf

        .. math::
            l(r) \\sigma_r(r) ^ 2 = 1/f(r) \\int_r^{\\infty} f(s) l(s) G M(s) / s^2 ds

        where l(r) is the 3d light profile
        M(s) is the enclosed 3d mass
        f is the solution to
        d ln(f)/ d ln(r) = 2 beta(r)

        :param r: 3d radius
        :param kwargs_mass: mass model parameters (following lenstronomy lens model conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to anisotropy type chosen.
            We refer to the Anisotropy() class for details on the parameters.
        :return: sigma_r**2
        """
        # l_r = self.light_profile.light_3d_interp(r, kwargs_light)
        l_r = self.lightProfile.light_3d(r, kwargs_light)
        f_r = self.anisotropy_solution(r, **kwargs_anisotropy)
        return (
            1
            / f_r
            / l_r
            * self._jeans_solution_integral(
                r, kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            * const.G
            / (const.arcsec * self.cosmo.dd * const.Mpc)
        )

    def mass_3d(self, r, kwargs):
        """Mass enclosed a 3d radius.

        :param r: in arc seconds
        :param kwargs: lens model parameters in arc seconds
        :return: mass enclosed physical radius in kg
        """
        mass_dimless = self._mass_profile.mass_3d(r, kwargs)
        mass_dim = (
            mass_dimless
            * const.arcsec**2
            * self.cosmo.dd
            * self.cosmo.ds
            / self.cosmo.dds
            * const.Mpc
            * const.c**2
            / (4 * np.pi * const.G)
        )
        return mass_dim

    def grav_potential(self, r, kwargs_mass):
        """Gravitational potential in SI units.

        :param r: radius (arc seconds)
        :param kwargs_mass:
        :return: gravitational potential
        """
        mass_dim = self.mass_3d(r, kwargs_mass)
        grav_pot = -const.G * mass_dim / (r * const.arcsec * self.cosmo.dd * const.Mpc)
        return grav_pot

    def draw_light(self, kwargs_light):
        """

        :param kwargs_light: keyword argument (list) of the light model
        :return: 3d radius (if possible), 2d projected radius, x-projected coordinate, y-projected coordinate
        """
        r = self.lightProfile.draw_light_3d(kwargs_light, n=1)[0]
        R, x, y = util.project2d_random(r)
        return r, R, x, y

    def delete_cache(self):
        """Delete interpolation function for a specific mass and light profile as well
        as for a specific anisotropy model.

        :return:
        """
        if hasattr(self, "_log_mass_3d"):
            del self._log_mass_3d
        if hasattr(self, "_interp_jeans_integral"):
            del self._interp_jeans_integral
        if hasattr(self, "_interp_I_R_sigma2"):
            del self._interp_I_R_sigma2
        self.lightProfile.delete_cache()
        self.delete_anisotropy_cache()

    def _I_R_sigma2(self, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Equation A15 in Mamon & Lokas 2005 as a logarithmic numerical integral (if
        option is chosen)

        :param R: 2d projected radius (in angular units)
        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. We refer to the Anisotropy() class for details on
            the parameters.
        :return: integral of A15 in Mamon&Lokas 2005
        """
        R = max(R, self._min_integrate)
        max_integrate = (
            self._max_integrate
        )  # make sure the integration of the Jeans equation is performed further out than the interpolation

        if self._log_int is True:
            min_log = np.log10(R)
            max_log = np.log10(max_integrate)
            dlogr = (max_log - min_log) / (self._interp_grid_num - 1)
            r_array = np.logspace(
                min_log + dlogr / 2.0, max_log + dlogr / 2.0, self._interp_grid_num
            )
            dlog_r = (np.log10(r_array[2]) - np.log10(r_array[1])) * np.log(10)

            IR_sigma2_ = self._integrand_A15(
                r_array, R, kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            IR_sigma2_dr = IR_sigma2_ * dlog_r * r_array
        else:
            r_array = np.linspace(
                start=R, stop=self._max_interpolate, num=self._interp_grid_num
            )
            dr = r_array[2] - r_array[1]

            IR_sigma2_ = self._integrand_A15(
                r_array + dr / 2.0, R, kwargs_mass, kwargs_light, kwargs_anisotropy
            )
            IR_sigma2_dr = IR_sigma2_ * dr

        IR_sigma2 = np.sum(IR_sigma2_dr)  # integral from angle to physical scales
        IR = self.lightProfile.light_2d_finite(R, kwargs_light)

        return IR_sigma2 * 2 * const.G / (const.arcsec * self.cosmo.dd * const.Mpc), IR

    def I_R_sigma2_and_IR(self, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Return I(R)*sigma^2 equation A15 in Mamon&Lokas 2005 as interpolation in log
        space, and I(R)

        :param R: projected radius
        :param kwargs_mass: mass profile keyword arguments
        :param kwargs_light: light model keyword arguments
        :param kwargs_anisotropy: stellar anisotropy keyword arguments
        :return: a tuple containing (I(R)*sigma^2, IR)
        """
        return self._I_R_sigma2_interp(R, kwargs_mass, kwargs_light, kwargs_anisotropy)

    def _I_R_sigma2_interp(self, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Equation A15 in Mamon&Lokas 2005 as interpolation in log space.

        :param R: projected radius
        :param kwargs_mass: mass profile keyword arguments
        :param kwargs_light: light model keyword arguments
        :param kwargs_anisotropy: stellar anisotropy keyword arguments
        :return: interpolated value of I(R)*sigma^2
        """
        R = np.maximum(R, self._min_integrate)

        if not hasattr(self, "_interp_I_R_sigma2"):
            min_log = np.log10(self._min_integrate)
            max_log = np.log10(self._max_integrate)
            R_array = np.logspace(
                min_log, max_log, self._interp_grid_num
            )  # self._interp_grid_num
            I_R_sigma2_array = []
            I_R_array = []
            for R_i in R_array:
                I_R_sigma2_, IR_ = self._I_R_sigma2(
                    R_i, kwargs_mass, kwargs_light, kwargs_anisotropy
                )
                I_R_sigma2_array.append(I_R_sigma2_)
                I_R_array.append(IR_)

            self._interp_I_R_sigma2 = interp1d(
                np.log(R_array), np.array(I_R_sigma2_array), fill_value="extrapolate"
            )
            self._interp_I_R = interp1d(
                np.log(R_array), np.array(I_R_array), fill_value="extrapolate"
            )
        return self._interp_I_R_sigma2(np.log(R)), self._interp_I_R(np.log(R))

    def _integrand_A15(self, r, R, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Integrand of A15 (in log space) in Mamon&Lokas 2005.

        :param r: 3d radius in arc seconds
        :param R: 2d projected radius
        :param kwargs_mass: mass model parameters (following lenstronomy lens model
            conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light
            model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to
            anisotropy type chosen. We refer to the Anisotropy() class for details on
            the parameters.
        :return: integrand
        """
        k_r = self.K(r, R, **kwargs_anisotropy)
        # l_r = self.light_profile.light_3d_interp(r, kwargs_light)
        # m_r = self._mass_3d_interp(r, kwargs_mass)
        l_r = self.lightProfile.light_3d(r, kwargs_light)
        m_r = self.mass_3d(r, kwargs_mass)
        out = k_r * l_r * m_r / r
        return out

    def _jeans_solution_integral(self, r, kwargs_mass, kwargs_light, kwargs_anisotropy):
        """Interpolated solution of the integral.

        .. math::
            \\int_r^{\\infty} f(s) l(s) G M(s) / s^2 ds

        :param r: 3d radius
        :param kwargs_mass: mass profile keyword arguments
        :param kwargs_light: light profile keyword arguments
        :param kwargs_anisotropy: anisotropy keyword arguments
        :return: interpolated solution of the Jeans integral
        (copped values at large radius as they become numerically inaccurate)
        """
        if not hasattr(self, "_interp_jeans_integral"):
            min_log = np.log10(self._min_integrate)
            # we extend the integral but ignore these outer solutions in the interpolation
            max_log = np.log10(self._max_integrate)
            r_array = np.logspace(min_log, max_log, self._interp_grid_num)
            dlog_r = (np.log10(r_array[2]) - np.log10(r_array[1])) * np.log(10)

            integrand_jeans = (
                self._integrand_jeans_solution(
                    r_array, kwargs_mass, kwargs_light, kwargs_anisotropy
                )
                * dlog_r
                * r_array
            )
            # flip array from inf to finite
            integral_jeans_r = np.cumsum(np.flip(integrand_jeans))
            # flip array back
            integral_jeans_r = np.flip(integral_jeans_r)
            # call 1d interpolation function
            self._interp_jeans_integral = interp1d(
                np.log(r_array[r_array <= self._max_integrate]),
                integral_jeans_r[r_array <= self._max_integrate],
                fill_value="extrapolate",
            )
        return self._interp_jeans_integral(np.log(r))

    def _integrand_jeans_solution(
        self, r, kwargs_mass, kwargs_light, kwargs_anisotropy
    ):
        """Integrand of A1 (in log space) in Mamon&Lokas 2005 to calculate the Jeans
        equation numerically f(s) l(s) M(s) / s^2.

        :param r: 3d radius
        :param kwargs_mass: mass model keyword arguments
        :param kwargs_light: light model keyword arguments
        :param kwargs_anisotropy: anisotropy model keyword argument
        :return: integrand value
        """
        f_r = self.anisotropy_solution(r, **kwargs_anisotropy)
        l_r = self.lightProfile.light_3d(r, kwargs_light)
        m_r = self._mass_3d_interp(r, kwargs_mass)
        out = f_r * l_r * m_r / r**2
        return out

    def _mass_3d_interp(self, r, kwargs, new_compute=False):
        """

        :param r: in arc seconds
        :param kwargs: lens model parameters in arc seconds
        :param new_compute: bool, if True, recomputes the interpolation
        :return: mass enclosed physical radius in kg
        """
        if not hasattr(self, "_log_mass_3d") or new_compute is True:
            r_array = np.logspace(
                np.log10(self._min_interpolate),
                np.log10(self._max_interpolate),
                self._interp_grid_num,
            )
            mass_3d_array = self.mass_3d(r_array, kwargs)
            mass_3d_array[mass_3d_array < 10.0 ** (-100)] = 10.0 ** (-100)
            self._log_mass_3d = interp1d(
                np.log(r_array),
                np.log(mass_3d_array / r_array),
                fill_value=(np.log(mass_3d_array[0] / r_array[0]), -1000),
                bounds_error=False,
            )
        return np.exp(self._log_mass_3d(np.log(r))) * np.minimum(
            r, self._max_interpolate
        )



__author__ = "xinchzhu"


from lenstronomy.LensModel.Profiles.gaussian_ellipse_kappa import (
    GaussianEllipseKappa,
)
from lenstronomy.LensModel.Profiles.multi_gaussian_ellipse_kappa import (
    MultiGaussianEllipseKappa,
)

import numpy as np
import numpy.testing as npt
import pytest


class TestGaussianEllipseKappa(object):
    """Test the Gaussian with Gaussian Ellipse Kappa."""

    """Including 2 options: fixed ellipticity (input one single value of each e1 and e2)
        /variable ellipticities (input lists for both e1 and e2 for different components)"""

    def setup_method(self):
        self.multi = MultiGaussianEllipseKappa()
        self.single = GaussianEllipseKappa()

    def test_function(self):
        ############
        # fixed ellipticity
        ############
        x, y = 1, 2
        amp = 1
        sigma = 1
        e1, e2 = 0.1, -0.1
        center_x, center_y = 1, 0
        f_ = self.multi.function(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        ############
        # variable ellipticities
        ############
        f_e = self.multi.function(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=[e1],
            e2=[e2],
            center_x=center_x,
            center_y=center_y,
        )
        f_single = self.single.function(
            x,
            y,
            amp=amp,
            sigma=sigma,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        npt.assert_almost_equal(f_, f_single, decimal=8)
        npt.assert_almost_equal(f_e, f_single, decimal=8)

    def test_derivatives(self):
        ############
        # fixed ellipticity
        ############
        x, y = 1, 2
        amp = 1
        sigma = 1
        e1, e2 = 0.1, -0.1
        center_x, center_y = 1, 0
        f_x, f_y = self.multi.derivatives(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        ############
        # variable ellipticities
        ############
        f_x_e, f_y_e = self.multi.derivatives(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=[e1],
            e2=[e2],
            center_x=center_x,
            center_y=center_y,
        )
        f_x_s, f_y_s = self.single.derivatives(
            x,
            y,
            amp=amp,
            sigma=sigma,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        npt.assert_almost_equal(f_x, f_x_s, decimal=8)
        npt.assert_almost_equal(f_y, f_y_s, decimal=8)
        npt.assert_almost_equal(f_x_e, f_x_s, decimal=8)
        npt.assert_almost_equal(f_y_e, f_y_s, decimal=8)

    def test_hessian(self):
        ############
        # fixed ellipticity
        ############
        x, y = 1, 2
        amp = 1
        sigma = 1
        e1, e2 = 0.1, -0.1
        center_x, center_y = 1, 0
        f_xx, f_xy, f_yx, f_yy = self.multi.hessian(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        ############
        # variable ellipticities
        ############
        f_xx_e, f_xy_e, f_yx_e, f_yy_e = self.multi.hessian(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=[e1],
            e2=[e2],
            center_x=center_x,
            center_y=center_y,
        )
        f_xx_s, f_xy_s, f_yx_s, f_yy_s = self.single.hessian(
            x,
            y,
            amp=amp,
            sigma=sigma,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        npt.assert_almost_equal(f_xx, f_xx_s, decimal=8)
        npt.assert_almost_equal(f_yy, f_yy_s, decimal=8)
        npt.assert_almost_equal(f_xy, f_xy_s, decimal=8)
        npt.assert_almost_equal(f_yx, f_yx_s, decimal=7)
        npt.assert_almost_equal(f_xx_e, f_xx_s, decimal=8)
        npt.assert_almost_equal(f_yy_e, f_yy_s, decimal=8)
        npt.assert_almost_equal(f_xy_e, f_xy_s, decimal=8)
        npt.assert_almost_equal(f_yx_e, f_yx_s, decimal=7)

    def test_density_2d(self):
        ############
        # fixed ellipticity
        ############
        x, y = 1, 2
        amp = 1
        sigma = 1
        e1, e2 = 0.1, -0.1
        center_x, center_y = 1, 0
        f_ = self.multi.density_2d(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        ############
        # variable ellipticities
        ############
        f_e = self.multi.density_2d(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=[e1],
            e2=[e2],
            center_x=center_x,
            center_y=center_y,
        )
        f_single = self.single.density_2d(
            x,
            y,
            amp=amp,
            sigma=sigma,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        npt.assert_almost_equal(f_, f_single, decimal=8)
        npt.assert_almost_equal(f_e, f_single, decimal=8)


if __name__ == "__main__":
    pytest.main()



from lenstronomy.LightModel.Profiles.linear import Linear, LinearEllipse
import numpy as np
import numpy.testing as npt


class TestLinear(object):
    def setup_method(self):
        self.linear = Linear()
        self.ellipse = LinearEllipse()

    def test_function(self):
        amp = 1
        k = 1
        x = np.array([0, 1])
        y = np.array([0, 0])
        flux_true = np.array([amp, amp + k])
        flux = self.linear.function(x, y, amp=amp, k=k)
        flux_ellipse = self.ellipse.function(x, y, amp=amp, k=k, e1=0, e2=0)
        npt.assert_almost_equal(flux_ellipse, flux, decimal=6)
        npt.assert_almost_equal(flux, flux_true, decimal=5)

    def test_total_flux(self):
        amp = 1
        k = 1
        total_flux = self.linear.total_flux(amp, k, center_x=0, center_y=0)
        total_flux_ellipse = self.ellipse.total_flux(amp, k, center_x=0, center_y=0)
        npt.assert_almost_equal(total_flux_ellipse, total_flux, decimal=6)



import numpy as np
from numpy.linalg import inv

__all__ = ["PositionLikelihood"]


class PositionLikelihood(object):
    """Likelihood of positions of multiply imaged point sources."""

    def __init__(
        self,
        point_source_class,
        image_position_uncertainty=0.005,
        astrometric_likelihood=False,
        image_position_likelihood=False,
        ra_image_list=None,
        dec_image_list=None,
        source_position_likelihood=False,
        source_position_tolerance=None,
        source_position_sigma=0.001,
        force_no_add_image=False,
        restrict_image_number=False,
        max_num_images=None,
    ):
        """

        :param point_source_class: Instance of PointSource() class
        :param image_position_uncertainty: uncertainty in image position uncertainty (1-sigma Gaussian radially),
         this is applicable for astrometric uncertainties as well as if image positions are provided as data
        :param astrometric_likelihood: bool, if True, evaluates the astrometric uncertainty of the predicted and modeled
         image positions with an offset 'delta_x_image' and 'delta_y_image'
        :param image_position_likelihood: bool, if True, evaluates the likelihood of the model predicted image position
         given the data/measured image positions
        :param ra_image_list: list or RA image positions per model component
        :param dec_image_list: list or DEC image positions per model component
        :param source_position_likelihood: bool, if True, ray-traces image positions back to source plane and evaluates
         relative errors in respect ot the position_uncertainties in the image plane (image_position_uncertainty)
        :param source_position_tolerance: tolerance level (in arc seconds in the source plane) of the different images.
         If set =! None, then the backwards ray tracing is performed on the images and demand on the same position of
         the source is meant to match the requirements, otherwise a punishing likelihood term is introduced
        :type source_position_tolerance: None or float
        :param source_position_sigma: r.m.s. value corresponding to a 1-sigma Gaussian likelihood accepted by the model
         precision in matching the source position transformed from the image plane
        :param force_no_add_image: bool, if True, will punish additional images appearing in the frame of the modelled
         image(first calculate them)
        :param restrict_image_number: bool, if True, searches for all appearing images in the frame of the data and
         compares with max_num_images
        :param max_num_images: integer, maximum number of appearing images. Default is the number of  images given in
         the Param() class
        """
        self._pointSource = point_source_class
        # TODO replace with public function of ray_shooting
        self._lensModel = point_source_class._lens_model
        self._astrometric_likelihood = astrometric_likelihood
        self._image_position_sigma = image_position_uncertainty
        self._source_position_sigma = source_position_sigma
        self._bound_source_position_tolerance = source_position_tolerance
        self._force_no_add_image = force_no_add_image
        self._restrict_number_images = restrict_image_number
        self._source_position_likelihood = source_position_likelihood
        self._max_num_images = max_num_images
        if max_num_images is None and restrict_image_number is True:
            raise ValueError(
                "max_num_images needs to be provided when restrict_number_images is True!"
            )
        self._image_position_likelihood = image_position_likelihood
        if ra_image_list is None:
            ra_image_list = []
        if dec_image_list is None:
            dec_image_list = []
        self._ra_image_list, self._dec_image_list = ra_image_list, dec_image_list

    def logL(self, kwargs_lens, kwargs_ps, kwargs_special, verbose=False):
        """

        :param kwargs_lens: lens model parameter keyword argument list
        :param kwargs_ps: point source model parameter keyword argument list
        :param kwargs_special: special keyword arguments
        :param verbose: bool
        :return: log likelihood of the optional likelihoods being computed
        """

        logL = 0
        if self._astrometric_likelihood is True:
            logL_astrometry = self.astrometric_likelihood(
                kwargs_ps, kwargs_special, self._image_position_sigma
            )
            logL += logL_astrometry
            if verbose is True:
                print("Astrometric likelihood = %s" % logL_astrometry)
        if self._force_no_add_image:
            additional_image_bool = self.check_additional_images(kwargs_ps, kwargs_lens)
            if additional_image_bool is True:
                logL -= 10.0**5
                if verbose is True:
                    print(
                        "force no additional image penalty as additional images are found!"
                    )
        if self._restrict_number_images is True:
            ra_image_list, dec_image_list = self._pointSource.image_position(
                kwargs_ps=kwargs_ps, kwargs_lens=kwargs_lens
            )
            if len(ra_image_list[0]) > self._max_num_images:
                logL -= 10.0**5
                if verbose is True:
                    print(
                        "Number of images found %s exceeded the limited number allowed %s"
                        % (len(ra_image_list[0]), self._max_num_images)
                    )
        if (
            self._source_position_likelihood is True
            or self._bound_source_position_tolerance is not None
        ):
            logL_source_pos = self.source_position_likelihood(
                kwargs_lens,
                kwargs_ps,
                self._source_position_sigma,
                hard_bound_rms=self._bound_source_position_tolerance,
                verbose=verbose,
            )
            logL += logL_source_pos
            if verbose is True:
                print("source position likelihood %s" % logL_source_pos)
        if self._image_position_likelihood is True:
            logL_image_pos = self.image_position_likelihood(
                kwargs_ps=kwargs_ps,
                kwargs_lens=kwargs_lens,
                sigma=self._image_position_sigma,
            )
            logL += logL_image_pos
            if verbose is True:
                print("image position likelihood %s" % logL_image_pos)
        return logL

    def check_additional_images(self, kwargs_ps, kwargs_lens):
        """Checks whether additional images have been found and placed in kwargs_ps.

        :param kwargs_ps: point source kwargs
        :param kwargs_lens: lens model keyword arguments
        :return: bool, True if more image positions are found than originally been
            assigned
        """
        ra_image_list, dec_image_list = self._pointSource.image_position(
            kwargs_ps=kwargs_ps, kwargs_lens=kwargs_lens, additional_images=True
        )
        for i in range(len(ra_image_list)):
            if "ra_image" in kwargs_ps[i]:
                if len(ra_image_list[i]) > len(kwargs_ps[i]["ra_image"]):
                    return True
        return False

    @staticmethod
    def astrometric_likelihood(kwargs_ps, kwargs_special, sigma):
        """Evaluates the astrometric uncertainty of the model plotted point sources
        (only available for 'LENSED_POSITION' point source model) and predicted image
        position by the lens model including an astrometric correction term.

        :param kwargs_ps: point source model kwargs list
        :param kwargs_special: kwargs list, should include the astrometric corrections
            'delta_x', 'delta_y'
        :param sigma: 1-sigma Gaussian uncertainty in the astrometry
        :return: log likelihood of the astrometirc correction between predicted image
            positions and model placement of the point sources
        """
        # TODO: make it compatible with multiple source instances
        if not len(kwargs_ps) > 0:
            return 0
        if "ra_image" not in kwargs_ps[0]:
            return 0
        if "delta_x_image" in kwargs_special:
            delta_x, delta_y = np.array(kwargs_special["delta_x_image"]), np.array(
                kwargs_special["delta_y_image"]
            )
            dist = (delta_x**2 + delta_y**2) / sigma**2 / 2
            logL = -np.sum(dist)
            if np.isnan(logL) is True:
                return -(10**15)
            return logL
        else:
            return 0

    def image_position_likelihood(self, kwargs_ps, kwargs_lens, sigma):
        """Computes the likelihood of the model predicted image position relative to
        measured image positions with an astrometric error. This routine requires the
        'ra_image_list' and 'dec_image_list' being declared in the initiation of the
        class.

        :param kwargs_ps: point source keyword argument list
        :param kwargs_lens: lens model keyword argument list
        :param sigma: 1-sigma uncertainty in the measured position of the images
        :return: log likelihood of the model predicted image positions given the
            data/measured image positions.
        """
        ra_image_list, dec_image_list = self._pointSource.image_position(
            kwargs_ps=kwargs_ps, kwargs_lens=kwargs_lens, original_position=True
        )
        logL = 0
        for i in range(
            len(ra_image_list)
        ):  # sum over the images of the different model components
            len_i = min(len(self._ra_image_list[i]), len(ra_image_list[i]))
            logL += -np.sum(
                (
                    (ra_image_list[i][:len_i] - self._ra_image_list[i][:len_i]) ** 2
                    + (dec_image_list[i][:len_i] - self._dec_image_list[i][:len_i]) ** 2
                )
                / sigma**2
                / 2
            )
        return logL

    def source_position_likelihood(
        self, kwargs_lens, kwargs_ps, sigma, hard_bound_rms=None, verbose=False
    ):
        """Computes a likelihood/punishing factor of how well the source positions of
        multiple images match given the image position and a lens model. The likelihood
        level is computed in respect of a displacement in the image plane and transposed
        through the Hessian into the source plane.

        :param kwargs_lens: lens model keyword argument list
        :param kwargs_ps: point source keyword argument list
        :param sigma: 1-sigma Gaussian uncertainty in the image plane
        :param hard_bound_rms: hard bound deviation between the mapping of the images
            back to the source plane (in source frame)
        :param verbose: bool, if True provides print statements with useful information.
        :return: log likelihood of the model reproducing the correct image positions
            given an image position uncertainty
        """
        if len(kwargs_ps) < 1:
            return 0
        logL = 0
        source_x, source_y = self._pointSource.source_position(kwargs_ps, kwargs_lens)
        redshift_list = self._pointSource._redshift_list
        for k in range(len(kwargs_ps)):
            if (
                "ra_image" in kwargs_ps[k]
                and self._pointSource.point_source_type_list[k] == "LENSED_POSITION"
            ):
                x_image = kwargs_ps[k]["ra_image"]
                y_image = kwargs_ps[k]["dec_image"]
                self._lensModel.change_source_redshift(redshift_list[k])
                # calculating the individual source positions from the image positions
                k_list = self._pointSource.k_list(k)
                for i in range(len(x_image)):
                    if k_list is not None:
                        k_lens = k_list[i]
                    else:
                        k_lens = None
                    x_source_i, y_source_i = self._lensModel.ray_shooting(
                        x_image[i], y_image[i], kwargs_lens, k=k_lens
                    )
                    f_xx, f_xy, f_yx, f_yy = self._lensModel.hessian(
                        x_image[i], y_image[i], kwargs_lens, k=k_lens
                    )
                    A = np.array([[1 - f_xx, -f_xy], [-f_yx, 1 - f_yy]])
                    Sigma_theta = np.array([[1, 0], [0, 1]]) * sigma**2
                    Sigma_beta = image2source_covariance(A, Sigma_theta)
                    delta = np.array(
                        [source_x[k] - x_source_i, source_y[k] - y_source_i]
                    )
                    if hard_bound_rms is not None:
                        if delta[0] ** 2 + delta[1] ** 2 > hard_bound_rms**2:
                            if verbose is True:
                                print(
                                    "Image positions of image %s of model %s do not match to the same source position to the required "
                                    "precision. Achieved: %s, Required: %s."
                                    % (i, k, delta, hard_bound_rms)
                                )
                            logL -= 10**3
                    try:
                        Sigma_inv = inv(Sigma_beta)
                    except:
                        return -(10**15)
                    chi2 = delta.T.dot(Sigma_inv.dot(delta))
                    logL -= chi2 / 2
        return logL

    @property
    def num_data(self):
        """

        :return: integer, number of data points associated with the class instance
        """
        num = 0
        if self._image_position_likelihood is True:
            for i in range(
                len(self._ra_image_list)
            ):  # sum over the images of the different model components
                num += len(self._ra_image_list[i]) * 2
        return num


# Equation (13) in Birrer & Treu 2019
def image2source_covariance(A, Sigma_theta):
    """
    computes error covariance in the source plane
    A: Hessian lensing matrix
    Sigma_theta: image plane covariance matrix of uncertainties
    """
    ATSigma = np.matmul(A.T, Sigma_theta)
    return np.matmul(ATSigma, A)



import scipy.special as special
import numpy as np
import scipy
from lenstronomy.Util import param_util

__all__ = ["SersicUtil"]


class SersicUtil(object):
    _s = 0.00001

    def __init__(self, smoothing=_s, sersic_major_axis=False):
        """

        :param smoothing: smoothing scale of the innermost part of the profile (for numerical reasons)
        :param sersic_major_axis: boolean; if True, defines the half-light radius of the Sersic light profile along
         the semi-major axis (which is the Galfit convention)
         if False, uses the product average of semi-major and semi-minor axis as the convention
         (default definition for all light profiles in lenstronomy other than the Sersic profile)
        """
        self._smoothing = smoothing
        self._sersic_major_axis = sersic_major_axis

    def k_bn(self, n, Re):
        """Returns normalisation of the sersic profile such that Re is the half light
        radius given n_sersic slope."""
        bn = self.b_n(n)
        k = bn * Re ** (-1.0 / n)
        return k, bn

    def k_Re(self, n, k):
        """"""
        bn = self.b_n(n)
        Re = (bn / k) ** n
        return Re

    @staticmethod
    def b_n(n):
        """B(n) computation. This is the approximation of the exact solution to the
        relation, 2*incomplete_gamma_function(2n; b_n) = Gamma_function(2*n).

        :param n: the sersic index
        :return: b(n)
        """
        bn = 1.9992 * n - 0.3271
        bn = np.maximum(
            bn, 0.00001
        )  # make sure bn is strictly positive as a save guard for very low n_sersic
        return bn

    def get_distance_from_center(self, x, y, e1, e2, center_x, center_y):
        """Get the distance from the center of Sersic, accounting for orientation and
        axis ratio :param x:

        :param y:
        :param e1: eccentricity
        :param e2: eccentricity
        :param center_x: center x of sersic
        :param center_y: center y of sersic
        """

        if self._sersic_major_axis:
            phi_G, q = param_util.ellipticity2phi_q(e1, e2)
            x_shift = x - center_x
            y_shift = y - center_y
            cos_phi = np.cos(phi_G)
            sin_phi = np.sin(phi_G)
            xt1 = cos_phi * x_shift + sin_phi * y_shift
            xt2 = -sin_phi * x_shift + cos_phi * y_shift
            xt2difq2 = xt2 / (q * q)
            r = np.sqrt(xt1 * xt1 + xt2 * xt2difq2)
        else:
            x_, y_ = param_util.transform_e1e2_product_average(
                x, y, e1, e2, center_x, center_y
            )
            r = np.sqrt(x_**2 + y_**2)
        return r

    def _x_reduced(self, x, y, n_sersic, r_eff, center_x, center_y):
        """Coordinate transform to normalized radius :param x:

        :param y:
        :param center_x:
        :param center_y:
        :return:
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        if isinstance(r, int) or isinstance(r, float):
            r = max(self._s, r)
        else:
            r[r < self._s] = self._s
        x_reduced = (r / r_eff) ** (1.0 / n_sersic)
        return x_reduced

    def _alpha_eff(self, r_eff, n_sersic, k_eff):
        """Deflection angle at r_eff :param r_eff:

        :param n_sersic:
        :param k_eff:
        :return:
        """
        b = self.b_n(n_sersic)
        alpha_eff = (
            n_sersic
            * r_eff
            * k_eff
            * b ** (-2 * n_sersic)
            * np.exp(b)
            * special.gamma(2 * n_sersic)
        )
        return -alpha_eff

    def alpha_abs(self, x, y, n_sersic, r_eff, k_eff, center_x=0, center_y=0):
        """

        :param x:
        :param y:
        :param n_sersic:
        :param r_eff:
        :param k_eff:
        :param center_x:
        :param center_y:
        :return:
        """
        n = n_sersic
        x_red = self._x_reduced(x, y, n_sersic, r_eff, center_x, center_y)
        b = self.b_n(n_sersic)
        a_eff = self._alpha_eff(r_eff, n_sersic, k_eff)
        alpha = 2.0 * a_eff * x_red ** (-n) * (special.gammainc(2 * n, b * x_red))
        return alpha

    def d_alpha_dr(self, x, y, n_sersic, r_eff, k_eff, center_x=0, center_y=0):
        """

        :param x:
        :param y:
        :param n_sersic:
        :param r_eff:
        :param k_eff:
        :param center_x:
        :param center_y:
        :return:
        """
        _dr = 0.00001
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        alpha = self.alpha_abs(r, 0, n_sersic, r_eff, k_eff)
        alpha_dr = self.alpha_abs(r + _dr, 0, n_sersic, r_eff, k_eff)
        d_alpha_dr = (alpha_dr - alpha) / _dr
        return d_alpha_dr

    def density(self, x, y, n_sersic, r_eff, k_eff, center_x=0, center_y=0):
        """De-projection of the Sersic profile based on Prugniel & Simien (1997)
        :return:"""
        raise ValueError(
            "not implemented! Use a Multi-Gaussian-component decomposition."
        )

    def _total_flux(self, r_eff, I_eff, n_sersic):
        """Computes total flux of a round Sersic profile.

        :param r_eff: projected half light radius
        :param I_eff: surface brightness at r_eff (in same units as r_eff)
        :param n_sersic: Sersic index
        :return: integrated flux to infinity
        """
        bn = self.b_n(n_sersic)
        return (
            I_eff
            * r_eff**2
            * 2
            * np.pi
            * n_sersic
            * np.exp(bn)
            / bn ** (2 * n_sersic)
            * scipy.special.gamma(2 * n_sersic)
        )

    def total_flux(self, amp, R_sersic, n_sersic, e1=0, e2=0, **kwargs):
        """Computes analytical integral to compute total flux of the Sersic profile.

        :param amp: amplitude parameter in Sersic function (surface brightness at
            R_sersic
        :param R_sersic: half-light radius in semi-major axis
        :param n_sersic: Sersic index
        :param e1: eccentricity
        :param e2: eccentricity
        :return: Analytic integral of the total flux of the Sersic profile
        """
        # compute product average half-light radius
        if self._sersic_major_axis:
            phi_G, q = param_util.ellipticity2phi_q(e1, e2)
            # translate semi-major axis R_eff into product averaged definition for circularization
            r_eff = R_sersic * np.sqrt(q)
        else:
            r_eff = R_sersic
        return self._total_flux(r_eff=r_eff, I_eff=amp, n_sersic=n_sersic)

    def _R_stable(self, R):
        """Floor R_ at self._smoothing for numerical stability.

        :param R: radius
        :return: smoothed and stabilized radius
        """
        return np.maximum(self._smoothing, R)

    def _r_sersic(
        self, R, R_sersic, n_sersic, max_R_frac=1000.0, alpha=1.0, R_break=0.0
    ):
        """

        :param R: radius (array or float)
        :param R_sersic: Sersic radius (half-light radius)
        :param n_sersic: Sersic index (float)
        :param max_R_frac: maximum window outside which the mass is zeroed, in units of R_sersic (float)
        :return: kernel of the Sersic surface brightness at R
        """

        R_ = self._R_stable(R)
        R_sersic_ = self._R_stable(R_sersic)
        bn = self.b_n(n_sersic)
        R_frac = R_ / R_sersic_
        if isinstance(R_, int) or isinstance(R_, float):
            if R_frac > max_R_frac:
                result = 0
            else:
                exponent = -bn * (R_frac ** (1.0 / n_sersic) - 1.0)
                result = np.exp(exponent)
        else:
            R_frac_real = R_frac[R_frac <= max_R_frac]
            exponent = -bn * (R_frac_real ** (1.0 / n_sersic) - 1.0)
            result = np.zeros_like(R_)
            result[R_frac <= max_R_frac] = np.exp(exponent)
        return np.nan_to_num(result)



__author__ = "sibirrer"


from lenstronomy.LensModel.Profiles.tnfw import TNFW
from lenstronomy.LensModel.Profiles.tnfw_ellipse_potential import TNFWELLIPSEPotential
from lenstronomy.LensModel.Profiles.nfw_ellipse_potential import NFWEllipsePotential

import numpy as np
import numpy.testing as npt
import pytest


class TestNFWELLIPSE(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.tnfw = TNFW()
        self.nfw_e = NFWEllipsePotential()
        self.tnfw_e = TNFWELLIPSEPotential()

    def test_function(self):
        x = np.linspace(start=0.1, stop=10, num=10)
        y = np.linspace(start=0.1, stop=10, num=10)
        # test round case against TNFW
        kwargs_tnfw_e_round = {
            "Rs": 1,
            "alpha_Rs": 0.1,
            "r_trunc": 5,
            "e1": 0.0,
            "e2": 0,
        }
        kwargs_tnfw_round = {"Rs": 1, "alpha_Rs": 0.1, "r_trunc": 5}
        f_e = self.tnfw_e.function(x, y, **kwargs_tnfw_e_round)
        f_r = self.tnfw.function(x, y, **kwargs_tnfw_round)
        npt.assert_almost_equal(f_e, f_r, decimal=5)

        # test elliptical case with r_trunc -> infinity against NFW_ELLIPSE
        kwargs_tnfw_e = {
            "Rs": 1,
            "alpha_Rs": 0.1,
            "r_trunc": 500,
            "e1": 0.2,
            "e2": -0.01,
        }
        kwargs_nfw_e = {"Rs": 1, "alpha_Rs": 0.1, "e1": 0.2, "e2": -0.01}
        f_te = self.tnfw_e.function(x, y, **kwargs_tnfw_e)
        f_e = self.nfw_e.function(x, y, **kwargs_nfw_e)
        npt.assert_almost_equal(f_te, f_e, decimal=3)

    def test_derivatives(self):
        x = np.linspace(start=0.1, stop=10, num=10)
        y = np.linspace(start=0.1, stop=10, num=10)
        # test round case against TNFW
        kwargs_tnfw_e_round = {
            "Rs": 1,
            "alpha_Rs": 0.1,
            "r_trunc": 5,
            "e1": 0.0,
            "e2": 0,
        }
        kwargs_tnfw_round = {"Rs": 1, "alpha_Rs": 0.1, "r_trunc": 5}
        f_xe, f_ye = self.tnfw_e.derivatives(x, y, **kwargs_tnfw_e_round)
        f_xr, f_yr = self.tnfw.derivatives(x, y, **kwargs_tnfw_round)
        npt.assert_almost_equal(f_xe, f_xr, decimal=5)
        npt.assert_almost_equal(f_ye, f_yr, decimal=5)

        # test elliptical case with r_trunc -> infinity against NFW_ELLIPSE
        kwargs_tnfw_e = {
            "Rs": 1,
            "alpha_Rs": 0.1,
            "r_trunc": 500,
            "e1": 0.2,
            "e2": -0.01,
        }
        kwargs_nfw_e = {"Rs": 1, "alpha_Rs": 0.1, "e1": 0.2, "e2": -0.01}
        out_te = self.tnfw_e.derivatives(x, y, **kwargs_tnfw_e)
        out_e = self.nfw_e.derivatives(x, y, **kwargs_nfw_e)
        npt.assert_almost_equal(out_te, out_e, decimal=3)

    def test_hessian(self):
        x = np.linspace(start=0.1, stop=10, num=10)
        y = np.linspace(start=0.1, stop=10, num=10)
        # test round case against TNFW
        kwargs_tnfw_e_round = {
            "Rs": 1,
            "alpha_Rs": 0.1,
            "r_trunc": 5,
            "e1": 0.0,
            "e2": 0,
        }
        kwargs_tnfw_round = {"Rs": 1, "alpha_Rs": 0.1, "r_trunc": 5}
        out_e = self.tnfw_e.hessian(x, y, **kwargs_tnfw_e_round)
        out_r = self.tnfw.hessian(x, y, **kwargs_tnfw_round)
        npt.assert_almost_equal(out_e, out_r, decimal=4)

        # test elliptical case with r_trunc -> infinity against NFW_ELLIPSE
        kwargs_tnfw_e = {
            "Rs": 1,
            "alpha_Rs": 0.1,
            "r_trunc": 500,
            "e1": 0.2,
            "e2": -0.01,
        }
        kwargs_nfw_e = {"Rs": 1, "alpha_Rs": 0.1, "e1": 0.2, "e2": -0.01}
        out_te = self.tnfw_e.hessian(x, y, **kwargs_tnfw_e)
        out_e = self.nfw_e.hessian(x, y, **kwargs_nfw_e)
        npt.assert_almost_equal(out_te, out_e, decimal=3)

    def test_mass_3d_lens(self):
        with npt.assert_raises(ValueError):
            kwargs_tnfw_e = {
                "Rs": 1,
                "alpha_Rs": 0.1,
                "r_trunc": 5,
                "e1": 0.1,
                "e2": -0.02,
            }
            self.tnfw_e.mass_3d_lens(1, **kwargs_tnfw_e)

    def test_density_lens(self):
        with npt.assert_raises(ValueError):
            kwargs_tnfw_e = {
                "Rs": 1,
                "alpha_Rs": 0.1,
                "r_trunc": 5,
                "e1": 0.1,
                "e2": -0.02,
            }
            self.tnfw_e.density_lens(1, **kwargs_tnfw_e)


if __name__ == "__main__":
    pytest.main()
